from .adam import Adam
from .sgd import SGD, Momentum
from .adagrad import AdaGrad
from .adadelta import AdaDelta
from .rmsprop import RMSprop
from xs.optim.optimizer import Optimizer
import copy


def get_optimizer(optimizer, **kwargs):
    parameters = kwargs.pop('parameters', None)
    lr = kwargs.pop('lr', 0.001)
    weight_decay = kwargs.pop('weight_decay', 0.)
    if optimizer.__class__.__name__ == 'str':
        optimizer = optimizer.lower()
        if optimizer == 'sgd':
            return SGD(parameters=parameters, lr=lr, weight_decay=weight_decay)
        elif optimizer == 'adam':
            return Adam(parameters=parameters, lr=lr, weight_decay=weight_decay)
        elif optimizer == 'rmsprop':
            return RMSprop(parameters=parameters, lr=lr, weight_decay=weight_decay)
        elif optimizer == 'momentum':
            return Momentum(parameters=parameters, lr=lr, weight_decay=weight_decay)
        elif optimizer == 'adagrad':
            return AdaGrad(parameters=parameters, lr=lr, weight_decay=weight_decay)
        elif optimizer == 'adadelta':
            return AdaDelta(parameters=parameters, lr=lr, weight_decay=weight_decay)
    elif isinstance(optimizer, Optimizer):
        return copy.deepcopy(optimizer)
    else:
        raise ValueError('unknown optimizer type!')


from xs.utils.common import *
import xs.nn.functional as F
import xs.nn.td_functional
from xs.layers.base import Layer, Zeros


class _Loss(Layer):
    def __init__(self, reduction: str = 'mean', **kwargs):
        self.reduction = reduction
        super(_Loss, self).__init__(**kwargs)

    @overload
    def init_layer_out_tensor(self, inputs: F.Tensor, target: F.Tensor):
        if self._data is None:
            if self.reduction == 'mean' or self.reduction == 'sum':
                self._data = Zeros()((1, ))
            else:
                self._data = Zeros()(inputs.shape)
            self._data.add_in_bounds(inputs, target)
            self._data.to('static')
        elif inputs.shape[0] < self._data.shape_capacity[0]:
            self._data.slices(slice(None, inputs.shape[0], None))
        else:
            self._data.slices(None)

    @overload
    def forward(self, inputs: F.Tensor, target: F.Tensor) -> F.Tensor:
        self._data = self.call(inputs, target)
        return self._data

    @overload
    def call(self, inputs: F.Tensor, target: F.Tensor) -> F.Tensor:
        raise NotImplemented

    def calc_acc(self, y_pred: F.Tensor, y_true: F.Tensor) -> float:
        raise NotImplemented

    def calc_loss(self, y_pred: F.Tensor, y_true: F.Tensor) -> float:
        raise NotImplemented

    def metric(self, y_pred: F.Tensor, y_true: F.Tensor) -> Tuple[float, float]:
        acc = self.calc_acc(y_pred, y_true)
        loss = self.calc_loss(y_pred, y_true)
        return acc, loss


class MSELoss(_Loss):
    def calc_acc(self, y_pred: F.Tensor, y_true: F.Tensor) -> float:
        return 0.

    def call(self, inputs: F.Tensor, target: F.Tensor) -> F.Tensor:
        return F.mse_loss(inputs, target, self.reduction)


class MAELoss(_Loss):
    def calc_acc(self, y_pred: F.Tensor, y_true: F.Tensor):
        return 0.

    def call(self, inputs: F.Tensor, target: F.Tensor) -> F.Tensor:
        self._data = F.mae_loss(inputs, target, self.reduction, self._data)
        return self._data


class BCELoss(_Loss):
    def calc_acc(self, y_pred: F.Tensor, y_true: F.Tensor):
        pred = y_pred.eval >= 0.5
        return GLOBAL.np.mean(pred == y_true.eval)

    def call(self, inputs: F.Tensor, target: F.Tensor) -> F.Tensor:
        self._data = F.bce_loss(inputs, target, self.reduction, self._data)
        return self._data


class CrossEntropyLoss(_Loss):
    def __init__(self):
        super(CrossEntropyLoss, self).__init__()

    @overload
    def init_layer_out_tensor(self, inputs: F.Tensor, target: F.Tensor):
        if self._data is None or inputs.shape[0] > self._data.shape_capacity[0]:
            if self.reduction == 'mean' or self.reduction == 'sum':
                self._data = Zeros()((1,))
            else:
                self._data = Zeros()(inputs.shape)
            self._data.add_in_bounds(inputs, target)
            inputs.cache['softmax'] = F.Tensor(data=GLOBAL.np.empty_like(inputs.eval))
            self._data.to('static')
        elif inputs.shape[0] < self._data.shape_capacity[0]:
            self._data.slices(slice(None, inputs.shape[0], None))
        else:
            self._data.slices(slice(None, None, None))

        if inputs.shape[0] < inputs.cache['softmax'].shape_capacity[0]:
            inputs.cache['softmax'].slices(slice(None, inputs.shape[0], None))
        else:
            inputs.cache['softmax'].slices(slice(None, None, None))

    def calc_loss(self, y_pred: F.Tensor, y_true: F.Tensor) -> float:
        return xs.nn.td_functional.nll_loss(xs.nn.td_functional.log_softmax(y_pred.eval), y_true.eval, self.reduction)

    def calc_acc(self, y_pred: F.Tensor, y_true: F.Tensor):
        acc = (GLOBAL.np.argmax(y_pred.eval, axis=-1).ravel() == y_true.eval.ravel())
        return GLOBAL.np.mean(acc).tolist()

    def call(self, inputs: F.Tensor, target: F.Tensor) -> F.Tensor:
        self._data = F.cross_entropy(inputs, target, self.reduction, self._data)
        return self._data


def get_objective(objective):
    if objective.__class__.__name__ == 'str':
        objective = objective.lower()
        if objective in ['crossentropy', 'cross_entropy']:
            return CrossEntropyLoss()
        # elif objective in ['sparsecrossentropy', 'sparse_crossentropy', 'sparse_cross_entropy']:
        #     return SparseCrossEntropy()
        elif objective in ['bce', 'binary_cross_entropy', 'binary_crossentropy']:
            return BCELoss()
        elif objective in ['meansquarederror', 'mean_squared_error', 'mse']:
            return MSELoss()
        elif objective in ['meanabsoluteerror', 'mean_absolute_error', 'mae']:
            return MAELoss()
    elif isinstance(objective, _Loss):
        return copy.deepcopy(objective)
    else:
        raise ValueError('unknown objective type!')



from .grad_fn import *
from xs.core import __global as GLOBAL


# math opearations
def add(lhs: Tensor, rhs: Tensor, out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = lhs
    if out is None:
        out = Tensor(GLOBAL.np.empty_like(lhs.eval))
    out.requires_grad = lhs.requires_grad or rhs.requires_grad
    xs.nn.td_functional.add(lhs.eval, rhs.eval, out.eval)
    if out.is_dynamic:
        lhs.add_out_bounds(out)
        rhs.add_out_bounds(out)
        out.add_in_bounds(lhs, rhs)

    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        initialize_ops_grad(lhs, rhs)
        out.grad_fn = AddBackward
    return out


def sub(lhs: Tensor, rhs: Tensor, out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = lhs
    if out is None:
        out = Tensor(GLOBAL.np.empty_like(lhs.eval))
    out.requires_grad = lhs.requires_grad or rhs.requires_grad
    xs.nn.td_functional.sub(lhs.eval, rhs.eval, out.eval)
    if out.is_dynamic:
        lhs.add_out_bounds(out)
        rhs.add_out_bounds(out)
        out.add_in_bounds(lhs, rhs)

    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        initialize_ops_grad(lhs, rhs)
        out.grad_fn = SubBackward
    return out


def mul(lhs: Tensor, rhs: Tensor, out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = lhs
    if out is None:
        out = Tensor(GLOBAL.np.empty_like(lhs.eval))
    out.requires_grad = lhs.requires_grad or rhs.requires_grad
    xs.nn.td_functional.mul(lhs.eval, rhs.eval, out.eval)

    if out.is_dynamic:
        lhs.add_out_bounds(out)
        rhs.add_out_bounds(out)
        out.add_in_bounds(lhs, rhs)

    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        initialize_ops_grad(lhs, rhs)
        out.grad_fn = MultiplyBackward
    return out


def pow(x: Tensor, exp: Union[int, float], out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = x
    if out is None:
        out = Tensor(GLOBAL.np.empty_like(x.eval))
    out.requires_grad = x.requires_grad
    GLOBAL.np.power(x.eval, exp, out=x.eval)

    if out.is_dynamic:
        x.add_out_bounds(out)
        out.add_in_bounds(x)

    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        initialize_ops_grad(x)
        out.cache['exp'] = exp
        out.grad_fn = PowBackward
    return out


# 矩阵乘法
def mm(lhs: Tensor, rhs: Tensor, out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = lhs
    if out is None:
        out = Tensor(GLOBAL.np.empty((lhs.shape[0], rhs.shape[-1])))
    out.requires_grad = lhs.requires_grad or rhs.requires_grad
    xs.nn.td_functional.mm(lhs.eval, rhs.eval, out.eval)

    if out.is_dynamic:
        out.add_in_bounds(lhs, rhs)
        lhs.add_out_bounds(out)
        rhs.add_out_bounds(out)
    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        initialize_ops_grad(lhs, rhs)
        out.grad_fn = MMBackward
    return out


def addmm(b: Tensor, x1: Tensor, x2: Tensor, out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = x1
    requires_grad = x1.requires_grad or x2.requires_grad or b.requires_grad
    if out is not None:
        assert isinstance(out, Tensor)
        xs.nn.td_functional.mm(x1.eval, x2.eval, out.eval)
        xs.nn.td_functional.add(out.eval, b.eval, out.eval)
    else:
        out_data = xs.nn.td_functional.mm(x1.eval, x2.eval)
        xs.nn.td_functional.add(out_data, b.eval, out_data)
        out = Tensor(data=out_data)
    out.requires_grad = requires_grad
    if out.is_dynamic:
        out.add_in_bounds(x1, x2, b)
        b.add_out_bounds(out)
        x1.add_out_bounds(out)
        x2.add_out_bounds(out)
    if GLOBAL.COMPUTE_GRAD and requires_grad:
        initialize_ops_grad(b, x1, x2)
        out.grad_fn = AddmmBackward
    return out


def truediv(lhs: Tensor, rhs: Tensor, out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = lhs
    if out is None:
        out = Tensor(GLOBAL.np.empty_like(lhs.eval))
    out.requires_grad = lhs.requires_grad or rhs.requires_grad
    xs.nn.td_functional.div(lhs.eval, rhs.eval, out.eval)

    if out.is_dynamic:
        lhs.add_out_bounds(out)
        rhs.add_out_bounds(out)
        out.add_in_bounds(lhs, rhs)

    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        initialize_ops_grad(lhs, rhs)
        out.grad_fn = DivBackward
    return out


def floordiv(lhs: Tensor, rhs: Tensor, out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = lhs
    if out is None:
        out = Tensor(GLOBAL.np.empty_like(lhs.eval))
    out.requires_grad = lhs.requires_grad or rhs.requires_grad
    xs.nn.td_functional.floor_div(lhs.eval, rhs.eval, out.eval)

    if out.is_dynamic:
        lhs.add_out_bounds(out)
        rhs.add_out_bounds(out)
        out.add_in_bounds(lhs, rhs)

    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        initialize_ops_grad(lhs, rhs)
        out.grad_fn = DivBackward
    return out


def max(x: Tensor, axis: int = None, keepdims: bool = False, out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = x
    if out is not None:
        assert isinstance(out, Tensor)
        xs.nn.td_functional.max(x.eval, axis, keepdims, out.eval)
    else:
        out = Tensor(xs.nn.td_functional.max(x.eval, axis, keepdims))

    if out.is_dynamic:
        out.add_in_bounds(x)
        x.add_out_bounds(out)
    out.requires_grad = x.requires_grad
    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        out.cache['axis'] = axis
        out.grad_fn = MaxBackward
        initialize_ops_grad(x)
    return out


def maximum(x1: Union[int, float, ndarray, Tensor], x2: Union[int, float, GLOBAL.np.ndarray, Tensor], out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = x1
    if isinstance(x1, Tensor):
        if out is None:
            out = Tensor(GLOBAL.np.empty_like(x1.eval))
        if isinstance(x2, (int, float, ndarray)):
            xs.nn.td_functional.maximum(x1.eval, x2, out.eval)
            out.requires_grad = x1.requires_grad
        elif isinstance(x2, Tensor):
            xs.nn.td_functional.maximum(x1.eval, x2.eval, out.eval)
            out.requires_grad = x1.requires_grad or x2.requires_grad
        else:
            raise TypeError('unknown type for {}'.format(type(x2)))
    elif isinstance(x2, Tensor):
        if out is None:
            out = Tensor(GLOBAL.np.empty_like(x2.eval))
        if isinstance(x1, (int, float, GLOBAL.np.ndarray)):
            xs.nn.td_functional.maximum(x1, x2.eval, out.eval)
            out.requires_grad = x2.requires_grad
        elif isinstance(x1, Tensor):
            xs.nn.td_functional.maximum(x1.eval, x2.eval, out.eval)
            out.requires_grad = x1.requires_grad or x2.requires_grad
        else:
            raise TypeError('unknown type for {}'.format(type(x1)))
    else:
        assert isinstance(x2, (int, float, GLOBAL.np.ndarray)) and isinstance(x1, (int, float, GLOBAL.np.ndarray))
        if out is None:
            out = Tensor(GLOBAL.np.maximum(x1, x2))
        else:
            GLOBAL.np.maximum(x1, x2, out=out.eval)

    if out.is_dynamic:
        out.add_in_bounds(x1, x2)
        x1.add_out_bounds(out)
        x2.add_out_bounds(out)

    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        out.grad_fn = MaximumBackward
        initialize_ops_grad(x1, x2)
    return out


def sum(x: Tensor, axis: int = None, keepdims: bool = False, out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = x
    if out is not None:
        assert isinstance(out, Tensor)
        xs.nn.td_functional.sum(x.eval, axis, keepdims, out.eval)
    else:
        out = Tensor(xs.nn.td_functional.sum(x.eval, axis, keepdims))

    if out.is_dynamic:
        out.add_in_bounds(x)
        x.add_out_bounds(out)
    out.requires_grad = x.requires_grad
    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        out.cache['axis'] = axis
        out.grad_fn = SumBackward
        initialize_ops_grad(x)
    return out


def mean(x: Tensor, axis: int = None, keepdims: bool = False, out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = x
    if out is not None:
        assert isinstance(out, Tensor)
        xs.nn.td_functional.mean(x.eval, axis, keepdims, out.eval)
    else:
        out = Tensor(xs.nn.td_functional.mean(x.eval, axis, keepdims))
    out.requires_grad = x.requires_grad
    if out.is_dynamic:
        out.add_in_bounds(x)
        x.add_out_bounds(out)
    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        out.cache['axis'] = axis
        out.grad_fn = MeanBackward
        initialize_ops_grad(x)
    return out


def norm(x: Tensor, p: int = 2, axis: int = None, keepdims: bool = False, out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = x
    if out is not None:
        assert isinstance(out, Tensor)
        xs.nn.td_functional.norm(x.eval, p, axis, keepdims, out.eval)
    else:
        out = Tensor(xs.nn.td_functional.norm(x.eval, p, axis, keepdims))
    if out.is_dynamic:
        out.add_in_bounds(x)
        x.add_out_bounds(out)
    out.requires_grad = x.requires_grad
    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        out.grad_fn = None
        initialize_ops_grad(x)
    return out


def transpose(x: Tensor, out: Tensor = None):
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = x
    if out is None:
        out = Tensor(GLOBAL.np.empty_like(x.eval))
    out.eval = GLOBAL.np.transpose(x.eval)
    if out.is_dynamic:
        out.add_in_bounds(x)
        x.add_out_bounds(out)

    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        out.grad_fn = TransposeBackward
        initialize_ops_grad(x)
    return out


def relu(inputs: Tensor, inplace: bool = False, out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = inputs
    if inplace:
        xs.nn.td_functional.relu(inputs.eval, inputs.eval)
        return inputs
    if out is None:
        out = Tensor(data=GLOBAL.np.empty_like(inputs.eval))
    xs.nn.td_functional.relu(inputs.eval, out.eval)
    out.requires_grad = inputs.requires_grad
    if out.is_dynamic:
        out.add_in_bounds(inputs)
        inputs.add_out_bounds(out)
    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        out.grad_fn = ReLUBackward
        initialize_ops_grad(inputs)
    return out


def sigmoid(inputs: Tensor, out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = inputs
    if out is None:
        out = Tensor(data=GLOBAL.np.empty_like(inputs.eval))
    xs.nn.td_functional.sigmoid(inputs.eval, out.eval)
    if out.is_dynamic:
        out.add_in_bounds(inputs)
        inputs.add_out_bounds(out)
    out.requires_grad = inputs.requires_grad
    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        out.grad_fn = SigmoidBackward
        initialize_ops_grad(inputs)
    return out


def tanh(inputs: Tensor, out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = inputs
    if out is None:
        out = Tensor(data=GLOBAL.np.empty_like(inputs.eval))
    xs.nn.td_functional.tanh(inputs.eval, out.eval)
    if out.is_dynamic:
        out.add_in_bounds(inputs)
        inputs.add_out_bounds(out)
    out.requires_grad = inputs.requires_grad
    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        out.grad_fn = TanhBackward
        initialize_ops_grad(inputs)
    return out


def softmax(inputs: Tensor, out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = inputs
    if out is None:
        out = Tensor(data=GLOBAL.np.empty_like(inputs.eval))
    out.requires_grad = inputs.requires_grad
    xs.nn.td_functional.softmax(inputs.eval, out.eval)
    if 'softmax' in inputs.cache.keys():
        inputs.cache['softmax'] = out
    return out


def log_softmax(inputs: Tensor, out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = inputs
    if out is None:
        out = Tensor(data=GLOBAL.np.empty_like(inputs.eval))
    out.requires_grad = inputs.requires_grad
    xs.nn.td_functional.log_softmax(inputs.eval, out.eval)
    return out


def flatten(inputs: Tensor, start: int = 1, out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = inputs
    if out is None:
        out = Tensor(data=xs.nn.td_functional.flatten(inputs.eval, start))
    else:
        out.eval = xs.nn.td_functional.flatten(inputs.eval, start)
    out.requires_grad = inputs.requires_grad
    if out.is_dynamic:
        out.add_in_bounds(inputs)
        inputs.add_out_bounds(out)
    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        out.grad_fn = FlattenBackward
        initialize_ops_grad(inputs)
    return out


def embedding(inputs: Tensor, weight: Tensor, out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = inputs
    if out is not None:
        out.eval = weight.eval[inputs.eval.astype(GLOBAL.np.int)]
    else:
        out = Tensor(data=weight.eval[inputs.eval.astype(GLOBAL.np.int)])
    if out.is_dynamic:
        out.add_in_bounds(inputs, weight)
        inputs.add_out_bounds(out)
        weight.add_out_bounds(out)
    out.requires_grad = inputs.requires_grad or weight.requires_grad
    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        out.grad_fn = EmbeddingBackward
        initialize_ops_grad(inputs, weight)
    return out


def conv2d(inputs: Tensor, weight: Tensor, bias: Tensor = None, stride: Tuple = (1, 1), padding: int = 0,
            out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = inputs
    # before pad size
    batch_nums, n_c_prev, n_h_prev, n_w_prev = inputs.eval.shape
    # pad
    pad_data = GLOBAL.np.pad(inputs.eval, ((0, 0), (0, 0), (padding, padding), (padding, padding)), 'constant')
    out_channels, in_channels, kernel_h, kernel_w = weight.eval.shape
    # output size
    n_h = (n_h_prev - kernel_h + 2 * padding) // stride[0] + 1
    n_w = (n_w_prev - kernel_w + 2 * padding) // stride[1] + 1
    col = im2col(pad_data, n_h, n_w, kernel_h, kernel_w, stride)
    col_w = weight.eval.reshape(out_channels, -1).T
    requires_grad = inputs.requires_grad or weight.requires_grad or (bias and bias.requires_grad)
    if out is None:
        out = Tensor(data=GLOBAL.np.empty((batch_nums, out_channels, n_h, n_w), dtype=inputs.dtype))

    data = GLOBAL.np.dot(col, col_w)
    if bias is not None:
        GLOBAL.np.add(data, bias.eval, out=data)
    out.eval = GLOBAL.np.reshape(data, (batch_nums, n_h, n_w, -1)).transpose(0, 3, 1, 2)
    out.requires_grad = requires_grad
    if out.is_dynamic:
        out.add_in_bounds(inputs, weight, bias)
        inputs.add_out_bounds(out)
        weight.add_out_bounds(out)
        if bias is not None:
            bias.add_out_bounds(out)

    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        # store these for bp
        out.cache['col'] = col
        out.cache['stride'] = stride
        out.cache['padding'] = padding
        out.grad_fn = Conv2DBackward
        initialize_ops_grad(inputs, weight, bias)
    return out


def max_pool2d(inputs: Tensor, kernel_size: int = 2, stride: int = None, padding: int = 0, out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = inputs
    if stride is None:
        stride = (kernel_size, kernel_size)
    elif isinstance(stride, int):
        stride = (stride, stride)
    if padding != 0:
        data = GLOBAL.np.pad(inputs.eval, ((0, 0), (0, 0), (padding, padding), (padding, padding)), 'constant')
    else:
        data = inputs.eval
    # after padding's shape if padded
    n, c, h, w = data.shape
    out_h = (h - kernel_size) // stride[0] + 1
    out_w = (w - kernel_size) // stride[1] + 1
    if out is None:
        out = Tensor(GLOBAL.np.empty((n, c, out_h, out_w), dtype=inputs.dtype))
    out.requires_grad = inputs.requires_grad
    if kernel_size == stride:
        mode = 'reshape'
        x_reshaped = data.reshape((n, c, h // kernel_size, kernel_size, w // kernel_size, kernel_size))
        data = GLOBAL.np.max(x_reshaped, axis=3)
        GLOBAL.np.max(data, axis=4, out=out.eval)
        if GLOBAL.COMPUTE_GRAD and out.requires_grad:
            out.cache['x_reshaped'] = x_reshaped

    else:
        mode = 'im2col'
        col = im2col(data, out_h, out_w, kernel_size, kernel_size, stride)
        col = col.reshape(-1, kernel_size * kernel_size)
        pool_argmax = GLOBAL.np.argmax(col, axis=1)
        col = GLOBAL.np.max(col, axis=1)
        out.eval = GLOBAL.np.reshape(col, (n, out_h, out_w, c)).transpose(0, 3, 1, 2)
        if GLOBAL.COMPUTE_GRAD and out.requires_grad:
            out.cache['pool_argmax'] = pool_argmax
            out.cache['kernel_size'] = kernel_size
    if out.is_dynamic:
        out.add_in_bounds(inputs)
        inputs.add_out_bounds(out)

    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        out.cache['mode'] = mode
        out.cache['stride'] = stride
        out.cache['padding'] = padding
        out.grad_fn = Maxpool2DBackward
        initialize_ops_grad(inputs)
    return out


def channel_max_pool(inputs: Tensor, kernel_size: int = 2, stride: int = 1, padding: int = 0, out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = inputs
    if stride is None:
        stride = (kernel_size, kernel_size)
    elif isinstance(stride, int):
        stride = (stride, stride)
    if padding != 0:
        data = GLOBAL.np.pad(inputs.eval, ((0, 0), (padding, padding), (0, 0), (0, 0)), 'constant')
    else:
        data = inputs.eval

    n, c, h, w = data.shape
    if kernel_size == stride:
        mode = 'reshape'
        x_reshaped = data.reshape((n, c // kernel_size, kernel_size, h, w))
        outputs = Tensor(data=x_reshaped.max(axis=2), in_bounds=[inputs, ], requires_grad=inputs.requires_grad)
        if GLOBAL.COMPUTE_GRAD and outputs.requires_grad:
            outputs.cache['x_reshaped'] = x_reshaped

    else:
        mode = 'im2col'
        out_c = (c - kernel_size) // stride + 1
        col = GLOBAL.np.zeros((n, kernel_size, out_c, h, w))
        for y in range(kernel_size):
            y_max = y + stride * out_c
            col[:, y] = data[:, y: y_max: stride]

        pool_argmax = GLOBAL.np.argmax(col, axis=1)
        outputs = GLOBAL.np.max(col, axis=1).reshape((n, out_c, h, w))
        outputs = Tensor(data=outputs, in_bounds=[inputs], requires_grad=inputs.requires_grad)
        if GLOBAL.COMPUTE_GRAD and outputs.requires_grad:
            outputs.cache['pool_argmax'] = pool_argmax
            outputs.cache['kernel_size'] = kernel_size

    inputs.out_bounds.append(outputs)
    if GLOBAL.COMPUTE_GRAD and outputs.requires_grad:
        outputs.cache['mode'] = mode
        outputs.cache['pad_size'] = padding
        outputs.cache['stride'] = stride
        outputs.grad_fn = ChannelMaxpoolBackward
        initialize_ops_grad(inputs)
    return outputs


def avg_pool2d(inputs: Tensor, kernel_size: int, stride: int = None, padding: int = 0, out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = inputs
    if stride is None:
        stride = (kernel_size, kernel_size)
    elif isinstance(stride, int):
        stride = (stride, stride)
    if padding != 0:
        data = GLOBAL.np.pad(inputs.eval, ((0, 0), (0, 0), (padding, padding), (padding, padding)), 'constant')
    else:
        data = inputs.eval

    n, c, h, w = data.shape
    out_h = (h - kernel_size + 2 * padding) // stride[0] + 1
    out_w = (w - kernel_size + 2 * padding) // stride[1] + 1
    if out is None:
        out = Tensor(GLOBAL.np.empty((n, c, out_h, out_w), dtype=inputs.dtype))
    out.requires_grad = inputs.requires_grad
    if kernel_size == stride:
        mode = 'reshape'
        x_reshaped = data.reshape((n, c, h // kernel_size, kernel_size, w // kernel_size, kernel_size))
        GLOBAL.np.mean(x_reshaped, axis=(3, 5), out=out.eval)
        if GLOBAL.COMPUTE_GRAD and out.requires_grad:
            out.cache['reshaped_shape'] = x_reshaped.shape
    else:
        mode = 'im2col'
        col = im2col(data, out_h, out_w, kernel_size, kernel_size, stride)
        col = col.reshape(-1, kernel_size * kernel_size)
        pool_argmean = GLOBAL.np.array([range(col.shape[1])])
        col = GLOBAL.np.mean(col, axis=1)
        out.eval = GLOBAL.np.reshape(col, (n, out_h, out_w, c)).transpose(0, 3, 1, 2)
        if GLOBAL.COMPUTE_GRAD and out.requires_grad:
            out.cache['pool_argmean'] = pool_argmean
            out.cache['kernel_size'] = kernel_size
    if out.is_dynamic:
        out.add_in_bounds(inputs)
        inputs.add_out_bounds(out)

    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        out.cache['mode'] = mode
        out.cache['stride'] = stride
        out.cache['padding'] = padding
        out.grad_fn = Avgpool2DBackward
        initialize_ops_grad(inputs)
    return out


def channel_avg_pool(inputs: Tensor, kernel_size: int = 2, stride: int = 1, padding: int = 0, out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = inputs
    if stride is None:
        stride = kernel_size
    if padding != 0:
        data = GLOBAL.np.pad(inputs.eval, ((0, 0), (padding, padding), (0, 0), (0, 0)), 'constant')
    else:
        data = inputs.eval

    n, c, h, w = data.shape
    if kernel_size == stride:
        mode = 'reshape'
        x_reshaped = data.reshape((n, c // kernel_size, kernel_size, h, w))
        outputs = Tensor(data=x_reshaped.mean(axis=2), in_bounds=[inputs, ], requires_grad=inputs.requires_grad)
        if GLOBAL.COMPUTE_GRAD and outputs.requires_grad:
            outputs.cache['x_reshaped'] = x_reshaped

    else:
        mode = 'im2col'
        out_c = (c - kernel_size) // stride + 1
        col = GLOBAL.np.zeros((n, kernel_size, out_c, h, w))
        for y in range(kernel_size):
            y_max = y + stride * out_c
            col[:, y] = data[:, y: y_max: stride]

        pool_argmean = GLOBAL.np.array([range(col.shape[1])])
        outputs = GLOBAL.np.mean(col, axis=1).reshape((n, out_c, h, w))
        outputs = Tensor(data=outputs, in_bounds=[inputs, ], requires_grad=inputs.requires_grad)
        if GLOBAL.COMPUTE_GRAD and outputs.requires_grad:
            outputs.cache['pool_argmean'] = pool_argmean
            outputs.cache['kernel_size'] = kernel_size

    inputs.out_bounds.append(outputs)
    if GLOBAL.COMPUTE_GRAD and outputs.requires_grad:
        outputs.cache['mode'] = mode
        outputs.cache['pad_size'] = padding
        outputs.cache['stride'] = stride
        outputs.grad_fn = ChannelAvgpoolBackward
        initialize_ops_grad(inputs)
    return outputs


def pad2d(inputs: Tensor, padding: tuple, out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = inputs
    if out is None:
        n, c, h, w = inputs.shape
        out_h = h + 2 * padding[0]
        out_w = w + 2 * padding[1]
        out = Tensor(GLOBAL.np.empty((n, c, out_h, out_w), dtype=inputs.dtype))
    out.eval = GLOBAL.np.pad(inputs.eval, ((0, 0), (0, 0), (padding[0], padding[0]), (padding[1], padding[1])),
                           'constant')
    if out.is_dynamic:
        out.add_in_bounds(inputs)
        inputs.add_out_bounds(out)
    out.requires_grad = inputs.requires_grad
    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        out.cache['padding'] = padding
        out.grad_fn = Pad2DBackward
        initialize_ops_grad(inputs)
    return out


def dropout2d(inputs: Tensor, keep_prob: float = 0.5, out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = inputs
    if not GLOBAL.COMPUTE_GRAD:
        return inputs
    if out is None:
        out = Tensor(data=GLOBAL.np.empty_like(inputs.eval))
    random_tensor = None
    if GLOBAL.TRAINING:
        random_tensor = GLOBAL.np.random.binomial(n=1, p=keep_prob, size=inputs.eval.shape)
        GLOBAL.np.multiply(inputs.eval, random_tensor, out=out.eval)
        GLOBAL.np.divide(out.eval, keep_prob, out=out.eval)
    else:
        out.eval = inputs.eval
    if out.is_dynamic:
        out.add_in_bounds(inputs)
        inputs.add_out_bounds(out)
    out.requires_grad = inputs.requires_grad
    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        out.cache['mask'] = random_tensor
        out.cache['keep_prob'] = keep_prob
        out.grad_fn = Dropout2DBackward
        initialize_ops_grad(inputs)
    return out


def batch_norm(inputs: Tensor, gamma: Tensor, beta: Tensor, moving_mean: Tensor, moving_variance: Tensor, axis: int = 1,
               training: bool = True, epsilon: float = 1e-5, momentum: float = 0.9, out: Tensor = None) -> Tensor:

    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = inputs
    if out is None:
        out = Tensor(data=GLOBAL.np.empty_like(inputs.eval))
    axis_field = tuple([i for i in range(inputs.eval.ndim) if i != axis])
    sqrtvar = None
    normalized_x = None
    if training:
        # calc mean
        mean = GLOBAL.np.mean(inputs.eval, axis=axis_field, keepdims=True)
        # calc var
        var = GLOBAL.np.var(inputs.eval, axis=axis_field, keepdims=True)
        # x minus u
        xmu = GLOBAL.np.subtract(inputs.eval, mean)
        sqrtvar = GLOBAL.np.sqrt(var + epsilon)
        normalized_x = GLOBAL.np.divide(xmu, sqrtvar)

        GLOBAL.np.multiply(xs.nn.td_functional.expand_as(gamma.eval, normalized_x), normalized_x, out=out.eval)
        GLOBAL.np.add(out.eval, xs.nn.td_functional.expand_as(beta.eval, out.eval), out=out.eval)

        # update moving mean
        GLOBAL.np.multiply(momentum, moving_mean.eval, out=moving_mean.eval)
        GLOBAL.np.multiply(1 - momentum, mean, out=mean)
        GLOBAL.np.add(moving_mean.eval, mean.ravel(), out=moving_mean.eval)
        # update moving variance
        GLOBAL.np.multiply(momentum, moving_variance.eval, out=moving_variance.eval)
        GLOBAL.np.multiply(1 - momentum, var, out=var)
        GLOBAL.np.add(moving_variance.eval, var.ravel(), out=moving_variance.eval)

    else:
        # moving_variance_data = GLOBAL.np.add(moving_variance.eval, epsilon)
        # GLOBAL.np.sqrt(moving_variance_data, out=moving_variance_data)
        # GLOBAL.np.divide(gamma.eval, moving_variance_data, out=moving_variance_data)
        #
        # moving_mean_data = GLOBAL.np.multiply(moving_variance_data, moving_mean.eval)
        # GLOBAL.np.subtract(beta.eval, moving_mean_data, out=moving_mean_data)
        #
        # GLOBAL.np.multiply(inputs.eval, moving_variance_data, out=moving_variance_data)
        # GLOBAL.np.add(moving_variance_data, moving_mean_data, out=out.eval)
        moving_variance_data = GLOBAL.np.add(moving_variance.eval, epsilon)
        GLOBAL.np.sqrt(moving_variance_data, out=moving_variance_data)
        moving_mean_data = GLOBAL.np.subtract(inputs.eval, xs.nn.td_functional.expand_as(moving_mean.eval, inputs.eval))
        GLOBAL.np.divide(moving_mean_data, xs.nn.td_functional.expand_as(moving_variance_data, moving_mean_data), out=out.eval)
        # GLOBAL.np.multiply(gamma.eval, out.eval, out=out.eval)
        GLOBAL.np.multiply(xs.nn.td_functional.expand_as(gamma.eval, out.eval), out.eval, out=out.eval)
        # GLOBAL.np.add(beta.eval, out.eval, out=out.eval)
        GLOBAL.np.add(xs.nn.td_functional.expand_as(beta.eval, out.eval), out.eval, out=out.eval)

    out.requires_grad = inputs.requires_grad or gamma.requires_grad or beta.requires_grad
    if out.is_dynamic:
        out.add_in_bounds(inputs, gamma, beta)
        inputs.add_out_bounds(out)
        gamma.add_out_bounds(out)
        beta.add_out_bounds(out)

    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        out.cache['axis_field'] = axis_field
        out.cache['sqrtvar'] = sqrtvar
        out.cache['normalized_x'] = normalized_x
        out.grad_fn = BatchNormBackward
        initialize_ops_grad(inputs, gamma, beta)
    return out


def layer_norm(inputs: Tensor, gamma: Tensor, beta: Tensor, epsilon: float = 1e-10, out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = inputs
    if out is None:
        out = Tensor(data=GLOBAL.np.empty_like(inputs.eval))
    shape_field = tuple([i for i in range(1, inputs.eval.ndim)])
    # calc mean
    xmu = GLOBAL.np.mean(inputs.eval, axis=shape_field, keepdims=True)
    # calc var
    sqrtvar = GLOBAL.np.var(inputs.eval, axis=shape_field, keepdims=True)
    # x minus u
    GLOBAL.np.subtract(inputs.eval, xmu, out=xmu)
    GLOBAL.np.sqrt(sqrtvar + epsilon, out=sqrtvar)

    normalized_x = GLOBAL.np.divide(xmu, sqrtvar)
    GLOBAL.np.multiply(xs.nn.td_functional.expand_as(gamma.eval, out.eval), normalized_x, out=out.eval)
    GLOBAL.np.add(out.eval, xs.nn.td_functional.expand_as(beta.eval, out.eval), out=out.eval)
    if out.is_dynamic:
        out.add_in_bounds(inputs, gamma, beta)
        inputs.add_out_bounds(out)
        gamma.add_out_bounds(out)
        beta.add_out_bounds(out)

    out.requires_grad = inputs.requires_grad or gamma.requires_grad or beta.requires_grad

    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        out.cache['shape_field'] = shape_field
        out.cache['xmu'] = xmu
        out.cache['sqrtvar'] = sqrtvar
        out.cache['normalized_x'] = normalized_x
        out.grad_fn = LayerNormBackward
        initialize_ops_grad(inputs, gamma, beta)
    return out


def group_norm(inputs: Tensor, gamma: Tensor, beta: Tensor,  epsilon: float = 1e-5, groups: int = 16, out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = inputs
    if out is None:
        out = Tensor(data=GLOBAL.np.empty_like(inputs.eval))

    n, c, h, w = inputs.eval.shape
    shape_field = tuple([i for i in range(2, inputs.eval.ndim)])
    data = GLOBAL.np.reshape(inputs.eval, (n, groups, c // groups, h, w))
    xgmu = GLOBAL.np.mean(data, axis=shape_field, keepdims=True)
    sqrtvar = GLOBAL.np.var(data, axis=shape_field, keepdims=True)
    GLOBAL.np.subtract(data, xgmu, out=xgmu)
    GLOBAL.np.sqrt(sqrtvar + epsilon, out=sqrtvar)

    x_group_norm = GLOBAL.np.divide(xgmu, sqrtvar)
    x_norm = GLOBAL.np.reshape(x_group_norm, (n, c, h, w))
    GLOBAL.np.multiply(xs.nn.td_functional.expand_as(gamma.eval, out.eval), x_norm, out=out.eval)
    GLOBAL.np.add(out.eval, xs.nn.td_functional.expand_as(beta.eval, out.eval), out=out.eval)

    if out.is_dynamic:
        out.add_in_bounds(inputs, gamma, beta)
        inputs.add_out_bounds(out)
        gamma.add_out_bounds(out)
        beta.add_out_bounds(out)

    out.requires_grad = inputs.requires_grad or gamma.requires_grad or beta.requires_grad
    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        out.cache['groups'] = groups
        out.cache['xgmu'] = xgmu
        out.cache['sqrtvar'] = sqrtvar
        out.cache['x_norm'] = x_norm
        out.grad_fn = GroupNormBackward
        initialize_ops_grad(inputs, gamma, beta)
    return out


def view(inputs: Tensor, shape: Tuple, out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = inputs
    if out is None:
        out = Tensor(data=GLOBAL.np.reshape(inputs.eval, shape))
    else:
        out.eval = GLOBAL.np.reshape(inputs.eval, shape)
    if out.is_dynamic:
        out.add_in_bounds(inputs)
        inputs.add_out_bounds(out)
    out.requires_grad = inputs.requires_grad
    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        out.grad_fn = ViewBackward
        initialize_ops_grad(inputs)
    return out


def concat(variables: List[Tensor], axis: int, name: str = None):
    data = variables[0].eval
    requires_grad = variables[0].requires_grad
    for i in range(1, len(variables)):
        data = GLOBAL.np.concatenate((data, variables[i].eval), axis=axis)
        requires_grad = requires_grad or variables[i].requires_grad
    out = Tensor(data=data, name=name, requires_grad=requires_grad)
    if out.is_dynamic:
        out.add_in_bounds(*variables)
    return out


def mse_loss(x1: Tensor, x2: Tensor, reduction: str = 'mean', out: Tensor = None) -> Tensor:
    assert x1.shape == x2.shape
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = x1
    if out is None:
        if reduction == 'mean' or reduction == 'sum':
            out = Tensor(GLOBAL.np.empty((1, )))
        elif reduction == 'none':
            out = Tensor(GLOBAL.np.empty_like(x1.eval))
        else:
            raise TypeError('unknown type for reduction')
    if reduction == 'sum':
        loss_val = GLOBAL.np.power(GLOBAL.np.subtract(x1.eval, x2.eval), 2)
        GLOBAL.np.sum(loss_val, out=out.eval)
    elif reduction == 'mean':
        loss_val = GLOBAL.np.sum(GLOBAL.np.power(GLOBAL.np.subtract(x1.eval, x2.eval), 2))
        GLOBAL.np.divide(loss_val, 2, out=out.eval)
        GLOBAL.np.divide(out.eval, GLOBAL.np.prod(x2.shape), out=out.eval)
    else:
        GLOBAL.np.subtract(x1.eval, x2.eval, out=out.eval)
        GLOBAL.np.power(out.eval, 2, out=out.eval)
    if out.is_dynamic:
        out.add_in_bounds(x1, x2)
        x1.add_out_bounds(out)
        x2.add_out_bounds(out)
    out.requires_grad = x1.requires_grad or x2.requires_grad

    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        out.cache['reduction'] = reduction
        out.grad_fn = MSELossBackward
        initialize_ops_grad(x1, x2)
    return out


def mae_loss(x1: Tensor, x2: Tensor, reduction: str = 'mean', out: Tensor = None) -> Tensor:
    assert x1.shape == x2.shape
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = x1
    if out is None:
        if reduction == 'mean' or reduction == 'sum':
            out = Tensor(GLOBAL.np.empty((1,)))
        elif reduction == 'none':
            out = Tensor(GLOBAL.np.empty_like(x1.eval))
        else:
            raise TypeError('unknown type for reduction')
    if reduction == 'sum':
        loss_val = GLOBAL.np.absolute(GLOBAL.np.subtract(x1.eval, x2.eval))
        GLOBAL.np.sum(loss_val, out=out.eval)
    elif reduction == 'mean':
        loss_val = GLOBAL.np.power(GLOBAL.np.subtract(x1.eval, x2.eval), 2)
        GLOBAL.np.sum(loss_val, out=out.eval)
        GLOBAL.np.divide(out.eval, x2.shape[0], out=out.eval)
    else:
        GLOBAL.np.subtract(x1.eval, x2.eval, out=out.eval)
        GLOBAL.np.absolute(out.eval, out=out.eval)
    if out.is_dynamic:
        out.add_in_bounds(x1, x2)
        x1.add_out_bounds(out)
        x2.add_out_bounds(out)
    out.requires_grad = x1.requires_grad or x2.requires_grad
    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        out.cache['reduction'] = reduction
        out.grad_fn = MAEBackward
        initialize_ops_grad(x1, x2)
    return out


def bce_with_logits_loss(x1: Tensor, x2: Tensor, reduction: str = 'mean', out: Tensor = None) -> Tensor:
    assert x1.shape == x2.shape
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = x1
    if out is None:
        if reduction == 'mean' or reduction == 'sum':
            out = Tensor(GLOBAL.np.empty((1,)))
        elif reduction == 'none':
            out = Tensor(GLOBAL.np.empty_like(x1.eval))
        else:
            raise TypeError('unknown type for reduction')

    pred = xs.nn.td_functional.sigmoid(x1.eval)
    xs.nn.td_functional.bce_loss(pred, x2.eval, reduction, out.eval)
    if out.is_dynamic:
        out.add_in_bounds(x1, x2)
        x1.add_out_bounds(out)
        x2.add_out_bounds(out)
    out.requires_grad = x1.requires_grad or x2.requires_grad
    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        out.grad_fn = BCEWithLogitsLossBackward
        initialize_ops_grad(x1, x2)
    return out


def bce_loss(x1: Tensor, x2: Tensor, reduction: str = 'mean', out: Tensor = None) -> Tensor:
    assert x1.shape == x2.shape
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = x1
    if out is None:
        if reduction == 'mean' or reduction == 'sum':
            out = Tensor(GLOBAL.np.empty((1,)))
        elif reduction == 'none':
            out = Tensor(GLOBAL.np.empty_like(x1.eval))
        else:
            raise TypeError('unknown type for reduction')
    xs.nn.td_functional.bce_loss(x1.eval, x2.eval, reduction, out.eval)
    if out.is_dynamic:
        out.add_in_bounds(x1, x2)
        x1.add_out_bounds(out)
        x2.add_out_bounds(out)
    out.requires_grad = x1.requires_grad or x2.requires_grad
    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        out.cache['reduction'] = reduction
        out.grad_fn = BCELossBackward
        initialize_ops_grad(x1, x2)
    return out


def nll_loss(x1: Tensor, x2: Tensor, reduction: str = 'mean', out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = x1
    if out is None:
        if reduction == 'mean' or reduction == 'sum':
            out = Tensor(GLOBAL.np.empty((1,)))
        elif reduction == 'none':
            out = Tensor(GLOBAL.np.empty_like(x1.eval))
        else:
            raise TypeError('unknown type for reduction')
    x2.long_()
    xs.nn.td_functional.nll_loss(x1.eval, x2.eval, reduction, out.eval)
    if out.is_dynamic:
        out.add_in_bounds(x1, x2)
        x1.add_out_bounds(out)
        x2.add_out_bounds(out)
        out.is_leaf = True
    out.requires_grad = x1.requires_grad or x2.requires_grad
    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        out.cache['reduction'] = reduction
        out.grad_fn = NllLossBackward
        initialize_ops_grad(x1, x2)
    return out


def cross_entropy(x1: Tensor, x2: Tensor, reduction: str = 'mean', out: Tensor = None) -> Tensor:
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = x1
    if out is None:
        if reduction == 'mean' or reduction == 'sum':
            out = Tensor(GLOBAL.np.empty((1,)))
        elif reduction == 'none':
            out = Tensor(GLOBAL.np.empty_like(x1.eval))
        else:
            raise TypeError('unknown type for reduction')
    x2.long_()
    if 'softmax' not in x1.cache.keys():
        x1.cache['softmax'] = None
    softmaxed_x1 = softmax(x1, x1.cache['softmax'])
    log_softmaxed_x1_data = GLOBAL.np.log(softmaxed_x1.eval)
    xs.nn.td_functional.nll_loss(log_softmaxed_x1_data, x2.eval, reduction, out.eval)
    if out.is_dynamic:
        out.add_in_bounds(x1, x2)
        x1.add_out_bounds(out)
        x2.add_out_bounds(out)
        out.is_leaf = True
    out.requires_grad = x1.requires_grad or x2.requires_grad
    if GLOBAL.COMPUTE_GRAD and out.requires_grad:
        out.cache['reduction'] = reduction
        out.grad_fn = CrossEntropyLossBackward
        initialize_ops_grad(x1, x2)
    return out


def clip(inputs: Tensor, min_val: float, max_val: float, out: Tensor = None):
    if out is None:
        out = Tensor(GLOBAL.np.empty_like(inputs.eval))
    GLOBAL.np.clip(inputs.eval, min_val, max_val, out=out.eval)
    return out


def lstm(inputs: Tensor, weight: Tensor, bias: Tensor, units: int,
         recurrent_activations: List, activations: List,
         prev_a: Tensor = None, c: Tensor = None, tao_f: Tensor = None,
         tao_u: Tensor = None, tao_o: Tensor = None, c_tilde: Tensor = None,
         return_sequences: bool = False):
    if GLOBAL.INPUTS is None:
        GLOBAL.INPUTS = inputs
    batch_nums, time_steps, n_vec = inputs.eval.shape
    if prev_a is None:
        prev_a = Tensor(GLOBAL.np.zeros((batch_nums, time_steps, n_vec)))
    if c is None:
        c = Tensor(GLOBAL.np.zeros((batch_nums, time_steps, n_vec)))
    if tao_f is None:
        tao_f = Tensor(GLOBAL.np.zeros((batch_nums, time_steps - 1, n_vec)))
    if tao_u is None:
        tao_u = Tensor(GLOBAL.np.zeros((batch_nums, time_steps - 1, n_vec)))
    if tao_o is None:
        tao_o = Tensor(GLOBAL.np.zeros((batch_nums, time_steps - 1, n_vec)))
    if c_tilde is None:
        c_tilde = Tensor(GLOBAL.np.zeros((batch_nums, time_steps - 1, n_vec)))

    z = GLOBAL.np.zeros((batch_nums, time_steps, n_vec + units))
    for t in range(1, time_steps + 1):
        zt = GLOBAL.np.concatenate((prev_a.eval[:, t - 1, :], inputs.eval[:, t - 1, :]), axis=1)
        ot = zt.dot(weight.eval) + bias.eval
        f = recurrent_activations[3 * (t - 1)].forward(Tensor(ot[:, :units]))
        u = recurrent_activations[3 * (t - 1) + 1].forward(Tensor(ot[:, units: units * 2]))
        _c_tilde = activations[t - 1].forward(Tensor(ot[:, units * 2: units * 3]))
        o = recurrent_activations[3 * (t - 1) + 2].forward(Tensor(ot[:, units * 3:]))

        c_tilde.eval[:, t - 1, :] = _c_tilde.eval
        _c = f.eval * c.eval[:, t - 1, :] + u.eval * _c_tilde.eval

        prev_a.eval[:, t, :] = o.eval * GLOBAL.np.tanh(_c)

        tao_f.eval[:, t - 1, :] = f.eval
        tao_u.eval[:, t - 1, :] = u.eval
        tao_o.eval[:, t - 1, :] = o.eval
        c.eval[:, t, :] = _c
        z[:, t - 1, :] = zt
    if return_sequences:
        outputs = Tensor(data=prev_a.eval[:, 1:, :], in_bounds=[inputs, ],
                           requires_grad=inputs.requires_grad or weight.requires_grad or (bias and bias.requires_grad))
    else:
        outputs = Tensor(data=prev_a.eval[:, -1, :], in_bounds=[inputs, ],
                           requires_grad=inputs.requires_grad or weight.requires_grad or (bias and bias.requires_grad))
    outputs.parameters([weight, bias])
    inputs.out_bounds.append(outputs)
    if GLOBAL.COMPUTE_GRAD and outputs.requires_grad:
        outputs.add_cache('units', units)
        outputs.cache['time_steps'] = time_steps
        outputs.cache['recurrent_activations'] = recurrent_activations
        outputs.cache['activations'] = activations
        outputs.cache['prev_a'] = prev_a
        outputs.cache['c'] = c
        outputs.cache['tao_f'] = tao_f
        outputs.cache['tao_u'] = tao_u
        outputs.cache['tao_o'] = tao_o
        outputs.cache['c_tilde'] = c_tilde
        outputs.cache['return_sequences'] = return_sequences
        outputs.grad_fn = LstmBackward
        initialize_ops_grad(inputs, weight, bias)
    return outputs



from .base import *
from xs.nn.initializers import Initializer
from xs.layers.activators import get_activator
import xs.nn.functional as F


class Conv2D(Layer):
    def __init__(self, out_channels: int, kernel_size: Union[int, Tuple], use_bias: bool = True,
                 stride: Union[int, Tuple] = 1, padding: Union[int, str] = 0, activation: str = None, kernel_initializer: Union[str, Initializer] = 'Normal',
                 bias_initializer: Union[str, Initializer] = 'zeros', **kwargs):
        self.out_channels = out_channels
        self.kernel_size = self.__check_size(kernel_size)
        self.use_bias = use_bias
        self.stride = self.__check_size(stride)
        self.padding = padding
        self.activation = get_activator(activation) if activation is not None else None
        self.kernel_initializer = get_initializer(kernel_initializer)
        self.bias_initializer = get_initializer(bias_initializer)
        self.pad_size = padding
        super(Conv2D, self).__init__(**kwargs)

    def __check_size(self, kernel_size):
        if isinstance(kernel_size, (int, float)):
            return int(kernel_size), int(kernel_size)
        elif isinstance(kernel_size, (Tuple, List)):
            return kernel_size

    def init_params(self, input_shape: Tuple = None, *args):
        if input_shape is not None:
            self._input_shape = input_shape
        weight = F.Parameter(self.kernel_initializer((self.out_channels, self._input_shape[0], self.kernel_size[0], self.kernel_size[1]), requires_grad=True))
        self._parameters.append(weight)
        if self.use_bias:
            bias = F.Parameter(self.bias_initializer((1, self.out_channels), requires_grad=True))
            self._parameters.append(bias)
        else:
            self._parameters.append(None)

    def compute_output_shape(self, input_shape: Union[List, Tuple] = None) -> Union[List, Tuple]:
        assert len(input_shape) == 3
        n_C_prev, n_H_prev, n_W_prev = input_shape
        filter_h, filter_w = self.kernel_size
        if self.padding.__class__.__name__ == 'str':
            padding = self.padding.upper()
            if padding == 'SAME':
                n_H = n_H_prev
                n_W = n_W_prev
                pad_h = (self.stride[0] * (n_H_prev - 1) - n_H_prev + filter_h) // 2
                self.pad_size = pad_h
            elif padding == 'VALID':
                n_H = (n_H_prev - filter_h) // self.stride[0] + 1
                n_W = (n_W_prev - filter_w) // self.stride[1] + 1
                self.pad_size = 0
            else:
                raise TypeError('Unknown padding type!plz inputs SAME or VALID or an integer')
        else:
            assert isinstance(self.padding, int)
            n_H = (n_H_prev - filter_h + 2 * self.padding) // self.stride[0] + 1
            n_W = (n_W_prev - filter_w + 2 * self.padding) // self.stride[1] + 1
            self.pad_size = self.padding
        return self.out_channels, n_H, n_W

    def init_layer_out_tensor(self, x: F.Tensor = None):
        super(Conv2D, self).init_layer_out_tensor(x)
        if self.activation is not None:
            self.activation.compute_output_shape(self._shape)
            self.activation.init_layer_out_tensor(self._data)

    def call(self, x: F.Tensor, *args, **kwargs) -> F.Tensor:
        if self.use_bias:
            weight, bias = self._parameters
        else:
            weight, bias = self._parameters[0], None

        self._data = F.conv2d(x, weight, bias, self.stride, self.pad_size,  self._data)
        if self.activation is not None:
            return self.activation.forward(self._data)
        return self._data

    def backward(self, gradients: F.Tensor = None):
        if self.activation is not None:
            self.activation.backward()
        super().backward()



import xs.core.__global as GLOBAL
import xs.nn


def backward(outputs, inputs, retain_graph=False):
    graph = topological_sort(inputs, outputs)
    for node in reversed(graph):
        if node.grad_fn is not None:
            node.grad_fn(node)
        for l in node.next_layers:
            l.data = None
        if retain_graph or node.is_leaf or node.retain_grad():
            node.reset_(requires_grad=node.requires_grad, name=node.name, slices=node.slices(),
                        static_graph_tensor=node.is_static, retain_grad=node.retain_grad(), grad=node.grad)
        else:
            node.free_memory()
            node.reset_(requires_grad=node.requires_grad, name=node.name, slices=node.slices(),
                        static_graph_tensor=node.is_static)

    if retain_graph:
        GLOBAL.GRAPH = graph
    else:
        del GLOBAL.INPUTS, GLOBAL.OUTPUTS, GLOBAL.GRAPH, graph[:]
        GLOBAL.INPUTS = None
        GLOBAL.OUTPUTS = None
        GLOBAL.GRAPH = None


def topological_sort(ins, outs):
    topo = []
    visited = set()

    def build_topo(v):
        if ins is not None and v == ins:
            topo.append(v)
            return
        if v not in visited:
            visited.add(v)
            for child in v.in_bounds:
                if isinstance(child, xs.nn.Parameter) or child is None:
                    continue
                build_topo(child)
            topo.append(v)
    build_topo(outs)
    return topo


def reset_graph(graph):
    del graph[:]
    del GLOBAL.INPUTS, GLOBAL.OUTPUTS, GLOBAL.GRAPH
    GLOBAL.INPUTS = None
    GLOBAL.OUTPUTS = None
    GLOBAL.GRAPH = None




from .base import *


class Activation(Layer):
    def __init__(self, act_name: str = 'relu', **kwargs):
        self.activation = get_activator(act_name)
        super(Activation, self).__init__(**kwargs)

    def init_layer_out_tensor(self, x: F.Tensor = None):
        super().init_layer_out_tensor(x)
        self.activation._data = self._data

    def call(self, x, *args, **kwargs) -> F.Tensor:
        self._data = self.activation.call(x)
        return self._data


class ReLU(Layer):
    def __init__(self, inplace: bool = False):
        super().__init__()
        self.inplace = inplace

    def init_layer_out_tensor(self, x = None):
        if self.inplace:
            self._data = x
            self._data.cache['mask'] = np.empty(self._data.shape, dtype=np.bool)
        else:
            if self._data is None or x.shape[0] > self._data.shape_capacity[0]:
                self._data = Zeros()((x.shape[0],) + self.shape, requires_grad=self.trainable)
                self._data.to('static')
                self._data.add_in_bounds(x)
            elif x.shape[0] < self._data.shape_capacity[0]:
                self._data.slices(slice(None, x.shape[0], None))
            else:
                self._data.slices(slice(None, None, None))

    def call(self, x: F.Tensor, *args, **kwargs) -> F.Tensor:
        if self.inplace:
            self._data = x
            if GLOBAL.TRAINING and x.requires_grad:
                if 'grad_fn' not in self._data.cache.keys():
                    self._data.cache['grad_fn'] = []
                self._data.cache['grad_fn'].append(self._data.grad_fn)
                self._data.cache['mask'] = x.eval < 0
                self._data.grad_fn = F.ReLUBackward
        self._data = F.relu(x, self.inplace, self._data)
        if GLOBAL.TRAINING and x.requires_grad:
            self._data.cache['inplace'] = self.inplace
            initialize_ops_grad(x)
        return self._data

    def backward(self, gradients: F.Tensor = None):
        if gradients is not None:
            assert isinstance(gradients, F.Tensor) and gradients.shape == self._data.shape
            self._data.grad = gradients
        if self._data.grad_fn is not None:
            self._data.grad_fn(self._data)
            if not self.inplace:
                self._data.zero_grad()


class Sigmoid(Layer):
    def call(self, x: F.Tensor, *args, **kwargs) -> F.Tensor:
        self._data = F.sigmoid(x,  self._data)
        return self._data


class Tanh(Layer):
    def call(self, x: F.Tensor, *args, **kwargs) -> F.Tensor:
        self._data = F.tanh(x,  self._data)
        return self._data


def get_activator(activator):
    if activator.__class__.__name__ == 'str':
        activator = activator.lower()
        if activator == 'relu':
            return ReLU()
        elif activator == 'sigmoid':
            return Sigmoid()
        elif activator == 'tanh':
            return Tanh()
        return None
    elif isinstance(activator, Activation):
        return F.copy.deepcopy(activator)
    else:
        raise TypeError('unknown activator type!')



class DataSet:
    def __init__(self, *datas):
        self.datas = list(datas)

    def __len__(self):
        return len(self.datas[0])

    def __getitem__(self, item):
        ret_list = []
        for data in self.datas:
            ret_list.append(data[item])
        return ret_list



from xs.optim.optimizer import *


class AdaGrad(Optimizer):
    def __init__(self, parameters=None, lr=0.01, weight_decay=0., epsilon=1e-7):
        self.epsilon = epsilon
        self.ms = None
        super(AdaGrad, self).__init__(parameters=parameters, lr=lr, weight_decay=weight_decay)

    def step(self):
        if self.ms is None:
            self.ms = [GLOBAL.np.zeros_like(g.eval) for g in self._parameters]
        for i, (s, var)in enumerate(zip(self.ms, self._parameters)):
            if var.requires_grad:
                s += GLOBAL.np.power(var.grad.eval, 2)
                var.eval -= self.lr * var.grad.eval / GLOBAL.np.sqrt(s + self.epsilon)
                self.ms[i] = s
        super(AdaGrad, self).step()



from .linear import Dense, Flatten
from .convolutional import Conv2D
from .pooling import MaxPooling2D, AvgPooling2D
from .normalize import BatchNormalization, LayerNormalization, GroupNormalization
from .activators import ReLU, Sigmoid, Tanh, Activation
from .ops import Input, Reshape, Add, ZeroPadding2D



from xs.optim.optimizer import *


class RMSprop(Optimizer):
    def __init__(self, parameters=None, lr=0.001, weight_decay=0., rho=0.9, epsilon=1e-7):
        self.rho = rho
        self.epsilon = epsilon
        self.ms = None
        super(RMSprop, self).__init__(parameters=parameters, lr=lr, weight_decay=weight_decay)

    def step(self):
        if self.ms is None:
            # initialize
            self.ms = [GLOBAL.np.zeros_like(p.eval) for p in self._parameters]

        for i, (s, var) in enumerate(zip(self.ms, self._parameters)):
            if var.requires_grad:
                new_s = self.rho * s + (1 - self.rho) * GLOBAL.np.square(var.grad.eval)
                var.eval -= self.lr * var.grad.eval / GLOBAL.np.sqrt(new_s + self.epsilon)
                self.ms[i] = new_s

        super(RMSprop, self).step()


from xs.optim.optimizer import *


class Adam(Optimizer):
    def __init__(self, parameters=None, lr=0.001, weight_decay=0.0, beta1=0.9, beta2=0.999, epsilon=1e-8):
        self.beta1 = beta1
        self.beta2 = beta2
        self.epsilon = epsilon
        self.ms = None
        self.vs = None
        super(Adam, self).__init__(parameters=parameters, lr=lr, weight_decay=weight_decay)

    def step(self):
        self.iterations += 1
        if self.ms is None:
            # initialize
            self.ms = [GLOBAL.np.zeros_like(p.eval) for p in self._parameters]
        if self.vs is None:
            # initialize
            self.vs = [GLOBAL.np.zeros_like(p.eval) for p in self._parameters]

        for i, (v, m, var) in enumerate(zip(self.vs, self.ms, self._parameters)):
            if var.requires_grad:
                v = self.beta1 * v + (1 - self.beta1) * var.grad.eval
                m = self.beta2 * m + (1 - self.beta2) * GLOBAL.np.square(var.grad.eval)
                v_correct = v / (1 - pow(self.beta1, self.iterations))
                m_correct = m / (1 - pow(self.beta2, self.iterations))
                var.eval -= self.lr * (v_correct / (GLOBAL.np.sqrt(m_correct) + self.epsilon))
                self.ms[i] = m
                self.vs[i] = v



from .base import *
import xs.nn.functional as F


class MaxPooling2D(Layer):
    def __init__(self, kernel_size: int = 2, stride: int = None, padding: int = 0, **kwargs):
        self.kernel_size = kernel_size
        self.stride = self.__check_size(kernel_size) if stride is None else self.__check_size(stride)
        self.padding = padding
        self.__mode = 'reshape'
        super(MaxPooling2D, self).__init__(**kwargs)

    def compute_output_shape(self, input_shape: Union[List, Tuple] = None) -> Union[List, Tuple]:
        n_c, n_h_prev, n_w_prev = input_shape
        if self.kernel_size == self.stride:
            self.__mode = 'reshape'
        else:
            self.__mode = 'im2col'
        n_h, n_w = (n_h_prev - self.kernel_size + 2 * self.padding) // self.stride[0] + 1, (n_w_prev - self.kernel_size
                                                                                            + 2 * self.padding) // self.stride[1] + 1
        return n_c, n_h, n_w

    def __check_size(self, kernel_size):
        if isinstance(kernel_size, (int, float)):
            return int(kernel_size), int(kernel_size)
        elif isinstance(kernel_size, (Tuple, List)):
            return kernel_size

    def call(self, x: F.Tensor, *args, **kwargs) -> F.Tensor:
        self._data = F.max_pool2d(x, self.kernel_size, self.stride, self.padding,  self._data)
        return self._data


class AvgPooling2D(Layer):
    def __init__(self, kernel_size: int = 2, stride: int = None, padding: int = 0, **kwargs):
        self.kernel_size = kernel_size
        self.stride = self.__check_size(kernel_size) if stride is None else self.__check_size(stride)
        self.padding = padding
        super(AvgPooling2D, self).__init__(**kwargs)

    def __check_size(self, kernel_size):
        if isinstance(kernel_size, (int, float)):
            return int(kernel_size), int(kernel_size)
        elif isinstance(kernel_size, (Tuple, List)):
            return kernel_size

    def compute_output_shape(self, input_shape: Union[List, Tuple] = None) -> Union[List, Tuple]:
        n_c, n_h_prev, n_w_prev = input_shape
        n_h, n_w = (n_h_prev - self.kernel_size + 2 * self.padding) // self.stride[0] + 1, \
                   (n_w_prev - self.kernel_size + 2 * self.padding) // self.stride[1] + 1
        return n_c, n_h, n_w

    def call(self, x, *args, **kwargs) -> F.Tensor:
        self._data = F.avg_pool2d(x, self.kernel_size, self.stride, self.padding,  self._data)
        return self._data


class ChannelMaxPooling(Layer):
    def __init__(self, kernel_size: int = 2, stride: int = None, padding: int = 0, **kwargs):
        self.kernel_size = kernel_size
        self.stride = kernel_size if stride is None else stride
        self.padding = padding
        self.__mode = 'reshape'
        super(ChannelMaxPooling, self).__init__(**kwargs)
        self.__data = None

    def compute_output_shape(self, input_shape: Union[List, Tuple] = None) -> Union[List, Tuple]:
        n_c_prev, n_h, n_w = input_shape
        if self.kernel_size == self.stride:
            self.__mode = 'reshape'
        else:
            self.__mode = 'im2col'
        n_c = (n_c_prev - self.kernel_size + 2 * self.padding) / self.stride + 1
        return n_c, n_h, n_w

    def call(self, x: F.Tensor, *args, **kwargs) -> F.Tensor:
        self.__data = F.channel_max_pool(x, self.kernel_size, self.stride, self.padding,  self.__data)
        return self.__data


class ChannelAvgPooling(Layer):
    def __init__(self, kernel_size: int = 2, stride: int = None, padding: int = 0):
        self.kernel_size = kernel_size
        self.stride = kernel_size if stride is None else stride
        self.padding = padding
        self.__mode = 'reshape'
        super(ChannelAvgPooling, self).__init__()
        self.__data = None

    def compute_output_shape(self, input_shape: Union[List, Tuple] = None) -> Union[List, Tuple]:
        n_c_prev, n_h, n_w = input_shape
        if self.kernel_size == self.stride:
            self.__mode = 'reshape'
        else:
            self.__mode = 'im2col'
        n_c = (n_c_prev - self.kernel_size + 2 * self.padding) / self.stride + 1
        return n_c, n_h, n_w

    def call(self, x: F.Tensor, *args, **kwargs) -> F.Tensor:
        self.__data = F.channel_avg_pool(x, self.kernel_size, self.stride, self.padding,  self.__data)
        return self.__data


import xs.nn
from xs.optim import get_optimizer
from xs.utils.common import GLOBAL, List, Tuple
import xs.nn.functional as F
import xs.nn.objectives
from xs.utils.toolkit import ProgressBar, format_time, SummaryProfile
import time
import pickle
import xs.utils.data as data
from collections import defaultdict
import xs.core.autograd


class _Base:
    def __init__(self):
        self._parameters = set()

    def train(self):
        GLOBAL.IS_TRAINING = True

    def eval(self):
        GLOBAL.IS_TRAINING = False

    def parameters(self, parameters: List[F.Tensor] = None):
        if parameters is None:
            return self._parameters
        else:
            self._parameters = set(parameters)

    def to(self, dst: str):
        if dst == 'cuda':
            if not GLOBAL.USE_CUDA:
                GLOBAL.USE_CUDA = True
                GLOBAL.np = __import__('cupy')
            for parameter in self._parameters:
                parameter.cuda_()
        elif dst == 'cpu':
            if GLOBAL.USE_CUDA:
                GLOBAL.USE_CUDA = False
                GLOBAL.np = __import__('numpy')
            for parameter in self._parameters:
                parameter.cpu_()
        return self


class _Model(_Base):
    def __init__(self):
        super().__init__()
        self.loss = None
        self.optimizer = None
        self._graph = None

    def compile(self, optimizer, loss):
        if GLOBAL.USE_CUDA:
            for parameter in self._parameters:
                parameter.cuda_()

    def init_graph_out_tensor(self, x, y):
        raise NotImplemented

    def fit(self,
            x: GLOBAL.np.ndarray,
            y: GLOBAL.np.ndarray,
            batch_size: int = None,
            epochs: int = 1,
            verbose: int = 1,
            shuffle: bool = True,
            validation_data: Tuple[GLOBAL.np.ndarray] = None,
            validation_split: float = 0.,
            initial_epoch: int = 0,
            ) -> SummaryProfile:
        x = GLOBAL.np.asarray(x)
        y = GLOBAL.np.asarray(y)
        record_data_names = ['train_loss', 'train_acc']
        if validation_data is None and 0. < validation_split < 1.:
            split = int(x.shape[0] * validation_split)
            valid_x, valid_y = x[-split:], y[-split:]
            train_x, train_y = x[:-split], y[:-split]
            validation_data = (valid_x, valid_y)
            record_data_names.append('validation_loss')
            record_data_names.append('validation_acc')
        else:
            train_x, train_y = x, y

        train_dataset = data.DataSet(train_x, train_y)
        train_dataloader = data.DataLoader(train_dataset, batch_size, shuffle)
        v_acc, v_loss = 0., 0.
        with SummaryProfile(*record_data_names) as profile:
            for epoch in range(initial_epoch, epochs):
                epoch_start_time = time.time()
                if verbose != 0:
                    print('\033[1;31m Epoch[%d/%d]\033[0m' % (epoch + 1, epochs))
                progress_bar = ProgressBar(max_iter=len(train_x), verbose=verbose)
                for idx, (batch_x, batch_y) in enumerate(train_dataloader):
                    if GLOBAL.USE_CUDA:
                        batch_x.cuda_()
                        batch_y.cuda_()
                    else:
                        batch_x.cpu_()
                        batch_y.cpu_()
                    if idx == 0 or idx == len(train_dataloader) - 1:
                        self.init_graph_out_tensor(batch_x, batch_y)
                    self.train()
                    # reset trainable_variables grad
                    self.optimizer.zero_grad()
                    # forward
                    pred = self.forward(batch_x)
                    loss = self.loss.forward(pred, batch_y)
                    self.backward(loss)
                    self.optimizer.step()
                    epoch_time = time.time() - epoch_start_time
                    train_acc = self.loss.calc_acc(pred, batch_y)
                    profile.step('train_acc', train_acc)
                    profile.step('train_loss', loss.item())

                    if validation_data is not None:
                        valid_x, valid_y = validation_data
                        v_acc, v_loss = self.evaluate(valid_x, valid_y, batch_size=batch_size)
                        profile.step('validation_loss', v_loss)
                        profile.step('validation_acc', v_acc)

                    progress_bar.update(batch_x.shape[0])
                    progress_bar.console(verbose, epoch_time, loss.item(), train_acc, v_loss,
                                         v_acc)

                if verbose != 0:
                    print()

        return profile

    def __call__(self, x: F.Tensor, *args, **kwargs):
        out = self.forward(x)
        return out

    def forward(self, x, *args, **kwargs):
        raise NotImplemented

    def backward(self, loss: F.Tensor):
        if loss.grad is None:
            loss.zero_grad()
            loss.grad.fill_(1.)
        loss.grad_fn(loss)
        for layer in reversed(self._graph):
            layer.backward()

    def evaluate(self, x: GLOBAL.np.ndarray, y: GLOBAL.np.ndarray, batch_size: int = None):
        self.eval()
        x = GLOBAL.np.asarray(x)
        y = GLOBAL.np.asarray(y)
        if batch_size is not None:
            assert type(batch_size) is int
            val_dataset = data.DataSet(x, y)
            val_dataloader = data.DataLoader(val_dataset, batch_size)
            acc_list = []
            loss_list = []
            for idx, (batch_x, batch_y) in enumerate(val_dataloader):
                if idx == 0 or idx == len(val_dataloader) - 1:
                    self.init_graph_out_tensor(batch_x, batch_y)

                y_pred = self.forward(batch_x)
                metric = self.loss.metric(y_pred, batch_y)
                acc_list.append(metric[0])
                loss_list.append(metric[1])

            acc = GLOBAL.np.array(acc_list).mean().tolist()
            loss = GLOBAL.np.array(loss_list).mean().tolist()
        else:
            y_pred = self.forward(F.Tensor(x))
            acc, loss = self.loss.metric(y_pred, F.Tensor(y))
        return acc, loss

    def predict(self, x: GLOBAL.np.ndarray, batch_size: int = None):
        self.eval()
        if batch_size is not None:
            assert type(batch_size) is int
            test_dataset = data.DataSet(x)
            test_dataloader = data.DataLoader(test_dataset, batch_size)
            pred_list = []
            for batch_x in test_dataloader:
                y_pred = self.forward(batch_x)
                pred_list.append(y_pred)
            pred = F.concat(pred_list, axis=0)
        else:
            pred = self.forward(F.Tensor(x))

        return pred

    def save(self, save_path):
        if not save_path.endswith('.pkl'):
            save_path += '.pkl'
        with open(save_path, 'wb') as f:
            pickle.dump([self._graph, self.optimizer, self.loss], f)

    def load(self, model_path):
        if not model_path.endswith('.pkl'):
            model_path += '.pkl'
        with open(model_path, 'rb') as f:
            graph, optimizer, loss = pickle.load(f)

        self._graph = graph
        self.optimizer = optimizer
        self.loss = loss

    def __str__(self):
        bar_nums = 75
        print('*' * bar_nums)
        print('Layer(type)'.ljust(25), 'Output Shape'.ljust(20), 'Param'.ljust(10), 'Connected to'.ljust(15))
        print('#' * bar_nums)
        if self._graph is None:
            raise ValueError('Please compile Model!')
        simple_name_dict = {
            'Conv2D': 'conv2d',
            'Input': 'input',
            'Dense': 'dense',
            'MaxPooling2D': 'max_pool2d',
            'AvgPooling2D': 'avg_pool2d',
            'Activation': 'activation',
            'ZeroPadding2D': 'zero_pad2d',
            'Add': 'add',
            'Flatten': 'flat'
        }
        simple_name_count_dict = defaultdict(int)
        for layer in self._graph:
            if layer.name is None:
                layer.name = simple_name_dict[layer.__class__.__name__] + str(simple_name_count_dict[simple_name_dict[layer.__class__.__name__]])
                simple_name_count_dict[simple_name_dict[layer.__class__.__name__]] += 1
            layer_name = '%s (%s)' % (layer.name, layer.__class__.__name__)
            params = layer.params_count()
            first = True
            if layer.in_bounds:
                for prev_layer in layer.in_bounds:
                    connected = prev_layer.name
                    if first:
                        print(layer_name.ljust(25), str((None,) + layer.shape).ljust(20), str(params).ljust(10),
                              connected.ljust(15))
                        first = False
                    else:
                        print(''.ljust(25), ''.ljust(20), ''.ljust(10), connected.ljust(15))
            else:
                connected = '\n'
                print(layer_name.ljust(25), str((None,) + layer.shape).ljust(20), str(params).ljust(10),
                      connected.ljust(15))
            print('-' * bar_nums)

        print('*' * bar_nums)
        total_params = 0
        trainable_params = 0
        for v in self._parameters:
            total_params += v.eval.size
            if v.requires_grad:
                trainable_params += v.eval.size
        params_details = 'Total params: %d\n' % total_params
        params_details += 'Trainable params: %d\n' % trainable_params
        params_details += 'Non-trainable params: %d\n' % (total_params - trainable_params)
        return params_details


class Sequential(_Model):
    def __init__(self, *layers):
        super().__init__()
        self._graph = [] if len(layers) == 0 else list(layers)
        self.__initialized = False

    def add(self, layer):
        self._graph.append(layer)

    def init_graph_out_tensor(self, x, y):
        for layer in self._graph:
            layer.init_layer_out_tensor(x)
            x = layer.data
        self.loss.init_layer_out_tensor(x, y)

    def init_params(self, input_shape=None):
        from xs.layers.base import Layer
        for layer in self._graph:
            if isinstance(layer, Layer):
                if len(layer.parameters()) == 0:
                    layer.init_params(input_shape)
                for v in layer.parameters():
                    if v is not None:
                        self._parameters.add(v)
                input_shape = layer.compute_output_shape(input_shape)
            self.__initialized = True

    def compile(self, optimizer, loss, **kwargs):
        assert self._graph
        next_layer = None
        for layer in self._graph:
            layer.connect(next_layer)
            next_layer = layer
        self.init_params()
        self.loss = xs.nn.objectives.get_objective(loss)
        self.optimizer = get_optimizer(optimizer, parameters=self._parameters, **kwargs)
        super(Sequential, self).compile(optimizer, loss)

    def forward(self, x, *args, **kwargs):
        if not self.__initialized:
            self.init_params(x.shape[1:])
        for layer in self._graph:
            x = layer.forward(x)
        return x

    def pop(self, index):
        layer = self._graph.pop(index)
        print('success delete %s layer' % layer.__class__.__name__)
        del layer


class Model(_Model):
    def __init__(self, inputs=None, outputs=None):
        super().__init__()
        self.inputs = inputs
        self.outputs = outputs

    def init_graph_out_tensor(self, x, y):
        self.inputs._data = x
        for i in range(1, len(self._graph)):
            self._graph[i].init_layer_out_tensor()
        self.loss.init_layer_out_tensor(self._graph[-1].data, y)

    def compile(self, optimizer, loss, **kwargs):
        assert self.inputs is not None and self.outputs is not None
        self._graph = xs.core.autograd.topological_sort(self.inputs, self.outputs)
        for g in self._graph:
            g.init_params()
            for v in g.parameters():
                if v is not None:
                    self._parameters.add(v)
        self.loss = xs.nn.objectives.get_objective(loss)
        self.optimizer = get_optimizer(optimizer, parameters=self._parameters, **kwargs)
        super(Model, self).compile(optimizer, loss)

    def forward(self, x: F.Tensor, *args, **kwargs):
        self.inputs._in_bounds = [x]
        outputs = None
        for layer in self._graph:
            outputs = layer.forward()
        return outputs


class Module(_Base):
    def __call__(self, x: F.Tensor, *args, **kwargs):
        out = self.forward(x)
        self.__collect_variables(x)
        out.is_leaf = True
        return out

    def __collect_variables(self, x: F.Tensor):
        queue = [x]
        seen = set()
        seen.add(x)
        while queue:
            vertex = queue.pop(0)
            for n in vertex.out_bounds:
                if n not in seen:
                    for v in n.in_bounds:
                        if isinstance(v, F.Parameter):
                            self._parameters.add(v)
                    queue.append(n)
                    seen.add(n)

    def forward(self, x):
        raise NotImplemented



from xs.layers.activators import get_activator
from xs.nn.initializers import get_initializer
from .base import *
import xs.nn.functional as F


class Dropout(Layer):
    def __init__(self, keep_prob: float, **kwargs):
        # prob :probability of keeping a unit active.
        self.keep_prob = keep_prob
        self.__data = None
        super(Dropout, self).__init__(**kwargs)

    def call(self, x: F.Tensor, *args, **kwargs) -> F.Tensor:
        self.__data = F.dropout2d(x, self.keep_prob,  self.__data)
        return self.__data


class BatchNormalization(Layer):
    def __init__(self, epsilon: float = 1e-6, momentum: float = 0.99, axis: int = 1, gamma_initializer: str = 'ones',
                 beta_initializer: str = 'zeros', moving_mean_initializer: str = 'zeros',
                 moving_variance_initializer: str = 'ones', **kwargs):
        # axis=1 when input Fully Connected Layers(data shape:(M,N),where M donotes Batch-size,and N represents feature nums)  ---also axis=-1 is the same
        # axis=1 when input Convolution Layers(data shape:(M,C,H,W),represents Batch-size,Channels,Height,Width,respectively)
        self.epsilon = epsilon
        self.axis = axis
        self.momentum = momentum
        self.gamma_initializer = get_initializer(gamma_initializer)
        self.beta_initializer = get_initializer(beta_initializer)
        self.moving_mean_initializer = get_initializer(moving_mean_initializer)
        self.moving_variance_initializer = get_initializer(moving_variance_initializer)
        self.moving_mean = None
        self.moving_variance = None
        super(BatchNormalization, self).__init__(**kwargs)

    def init_params(self, input_shape: Tuple = None, *args):
        if input_shape is not None:
            self._input_shape = input_shape
        assert len(self._input_shape) >= 1
        n_in = self._input_shape[self.axis - 1]
        gamma = F.Parameter(self.gamma_initializer(n_in, requires_grad=True))
        beta = F.Parameter(self.beta_initializer(n_in, requires_grad=True))
        self.moving_mean = self.moving_mean_initializer(n_in)
        self.moving_variance = self.moving_variance_initializer(n_in)
        self._parameters.append(gamma)
        self._parameters.append(beta)

    def call(self, x: F.Tensor, *args, **kwargs) -> F.Tensor:
        gamma, beta = self._parameters
        self._data = F.batch_norm(x, gamma, beta, self.moving_mean, self.moving_variance, self.axis, GLOBAL.TRAINING,
                                  self.epsilon, self.momentum, self._data)
        return self._data


class LayerNormalization(Layer):
    def __init__(self, epsilon: float = 1e-10, gamma_initializer: str = 'ones', beta_initializer: str = 'zeros'):
        self.epsilon = epsilon
        self.gamma_initializer = get_initializer(gamma_initializer)
        self.beta_initializer = get_initializer(beta_initializer)
        super(LayerNormalization, self).__init__()

    def init_params(self, input_shape: Tuple = None, *args):
        if input_shape is not None:
            self._input_shape = input_shape
        gamma = F.Parameter(self.gamma_initializer(self._input_shape, requires_grad=True))
        beta = F.Parameter(self.beta_initializer(self._input_shape, requires_grad=True))
        self._parameters.append(gamma)
        self._parameters.append(beta)

    def call(self, x: F.Tensor, *args, **kwargs) -> F.Tensor:
        gamma, beta = self._parameters
        self._data = F.layer_norm(x, gamma, beta,
                                 self.epsilon, self._data)
        return self._data


class GroupNormalization(Layer):
    def __init__(self, epsilon: float = 1e-5, groups: int = 16, gamma_initializer: str = 'ones', beta_initializer: str = 'zeros'):
        self.epsilon = epsilon
        self.G = groups
        self.gamma_initializer = get_initializer(gamma_initializer)
        self.beta_initializer = get_initializer(beta_initializer)
        super(GroupNormalization, self).__init__()

    def init_params(self, input_shape: Tuple = None, *args):
        if input_shape is not None:
            self._input_shape = input_shape
        c = self._input_shape[0]
        assert c % self.G == 0
        gamma = F.Parameter(self.gamma_initializer(c, requires_grad=True))
        beta = F.Parameter(self.beta_initializer(c, requires_grad=True))
        self._parameters.append(gamma)
        self._parameters.append(beta)

    def call(self, x: F.Tensor, *args, **kwargs) -> F.Tensor:
        gamma, beta = self._parameters
        self._data = F.group_norm(x, gamma, beta,
                                 self.epsilon, self.G, self._data)
        return self._data



from xs.optim.optimizer import *


class AdaDelta(Optimizer):
    def __init__(self, parameters=None, lr=1.0, weight_decay=0.0, rho=0.95, epsilon=1e-7):
        self.rho = rho
        self.epsilon = epsilon
        self.ms = None
        self.delta_x = None
        super(AdaDelta, self).__init__(parameters=parameters, lr=lr, weight_decay=weight_decay)

    def step(self):
        if self.ms is None:
            self.ms = [GLOBAL.np.zeros_like(g.grad.eval) for g in self._parameters]
        if self.delta_x is None:
            self.delta_x = [GLOBAL.np.zeros_like(g.grad.eval) for g in self._parameters]

        for i, (s, var, x) in enumerate(zip(self.ms, self._parameters, self.delta_x)):
            if var.requires_grad:
                s = self.rho * s + (1 - self.rho) * GLOBAL.np.power(var.grad.eval, 2)
                g_ = GLOBAL.np.sqrt((x + self.epsilon) / (s + self.epsilon)) * var.grad.eval
                var.eval -= g_
                x = self.rho * x + (1 - self.rho) * GLOBAL.np.power(g_, 2)
                self.ms[i] = s
                self.delta_x[i] = x
        super(AdaDelta,self).step()



from xs.utils.common import *
from xs.nn.initializers import Zeros, get_initializer
import xs.nn.functional as F


class Layer:
    def __init__(self, trainable: bool = True, name: str = None, in_bounds: List = None,
                 input_shape: Union[List, Tuple] = None, shape: Union[List, Tuple] = None,
                 parameters: List[F.Tensor] = None,  **kwargs):
        self.trainable = trainable
        self._in_bounds = in_bounds if in_bounds is not None else []
        self._out_bounds = []
        self._input_shape = input_shape
        self.name = str(name)
        self._shape = shape
        self._parameters = parameters if parameters is not None else []
        self._data = None
        super(Layer, self).__init__(**kwargs)

    def __call__(self, x, *args, **kwargs):
        if isinstance(x, F.Tensor):
            if len(self._parameters) == 0:
                self.init_params(x.shape[1:])
            x.next_layers.append(self)
            return self.call(x, *args, **kwargs)
        # elif isinstance(x, Layer):
        else:
            self._shape = self.compute_output_shape(x.shape)
            self._in_bounds.append(x)
            self._input_shape = x.shape
            x.add_out_bounds(self)
            return self
        # else:
        #     raise TypeError('unknown type for {}'.format(x))

    def connect(self, inbound=None):
        if inbound is None:
            if self._input_shape is None:
                raise ValueError('must specify input_shape')
        else:
            self._input_shape = inbound.shape
        self._shape = self.compute_output_shape(self._input_shape)
        if inbound is not None:
            self._in_bounds.append(inbound)
            inbound.add_out_bounds(self)

    def params_count(self) -> int:
        total_params = 0
        for v in self._parameters:
            if v is not None:
                total_params += v.eval.size
        return total_params

    @property
    def shape(self):
        return self._shape

    @property
    def out_bounds(self):
        return self._out_bounds

    @property
    def in_bounds(self):
        return self._in_bounds

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, v):
        self._data = v

    def parameters(self, parameters: List[F.Tensor] = None):
        if parameters is None:
            return self._parameters
        else:
            self._parameters = parameters

    def add_out_bounds(self, *outs):
        for out in outs:
            if isinstance(out, Layer):
                self._out_bounds.append(out)

    def init_params(self, input_shape: Tuple = None, *args):
        pass
    
    def init_layer_out_tensor(self, x: F.Tensor = None):
        x = self._in_bounds[0].data if x is None else x
        if self._data is None or x.shape[0] > self._data.shape_capacity[0]:
            self._data = Zeros()((x.shape[0],) + self.shape, requires_grad=self.trainable)
            self._data.to('static')
            self._data.add_in_bounds(x)
            self._data.add_in_bounds(*self.parameters())
        elif x.shape[0] < self._data.shape_capacity[0]:
            if GLOBAL.TRAINING:
                self._data.slices(slice(None, x.shape[0], None))
            else:
                self._data = Zeros()((x.shape[0],) + self.shape, requires_grad=self.trainable)
                self._data.to('static')
                self._data.add_in_bounds(x)
                self._data.add_in_bounds(*self.parameters())
        else:
            self._data.slices(slice(None, None, None))

    def compute_output_shape(self, input_shape: Union[List, Tuple] = None) -> Union[List, Tuple]:
        if isinstance(input_shape, int):
            self._shape = (input_shape, )
        else:
            self._shape = tuple(input_shape)
        return self._shape

    def call(self, x: F.Tensor, *args, **kwargs) -> F.Tensor:
        raise NotImplemented

    def forward(self, x: F.Tensor = None, *args, **kwargs) -> F.Tensor:
        if x is None:
            x = self._in_bounds[-1]
        if len(self._parameters) == 0:
            self.init_params(x.shape[1:])
        if isinstance(x, F.Tensor):
            if x.is_dynamic:
                x.next_layers.append(self)
            return self.call(x, *args, **kwargs)
        else:
            return self.call(x.data, *args, **kwargs)

    def backward(self, gradients: F.Tensor = None):
        if gradients is not None:
            assert isinstance(gradients, F.Tensor) and gradients.shape == self._data.shape
            self._data.grad = gradients
        if self._data.grad_fn is not None:
            self._data.grad_fn(self._data)
            self._data.zero_grad()



from .common import no_grad
import warnings
from collections import OrderedDict
import time
import os


class SummaryProfile:
    def __init__(self, *names: str, **kwargs):
        self.__start_time = None
        self.__psutil = None
        self.__data = OrderedDict()
        for name in names:
            self.__data[name] = []
        self.__print_gap = kwargs.pop('print_gap', None)
        self.__steps = 0

    def __enter__(self):
        try:
            self.__psutil = __import__('psutil')
        except ImportError:
            warnings.warn('Please install psutil module!')
        self.__start_time = time.time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        print(f'\033[1;31mTime cost: {format_time(time.time() - self.__start_time)}\033[0m')
        if self.__psutil is not None:
            print('\033[1;31mMemory cost：%.4f GB\033[0m' % (self.__psutil.Process(os.getpid()).memory_info().rss / 1024 / 1024 / 1024))

    def step_all(self, *datas):
        for i, k in enumerate(self.__data.keys()):
            self.__data[k].append(datas[i])
            if self.__print_gap is not None and self.__steps % self.__print_gap == 0:
                if i == 0:
                    print('### Step', self.__steps, end=' ### ')
                print(k, end=': ')
                if i == len(self.__data) - 1:
                    print(datas[i])
                else:
                    print(datas[i], end=', ')
        self.__steps += 1

    def step(self, name, value):
        self.__data[name].append(value)

    def visualize(self):
        import matplotlib.pyplot as plt
        for name, value in self.__data.items():
            plt.figure()
            plt.plot(value)
            plt.title(name)
        plt.show()

    @property
    def data(self) -> OrderedDict:
        return self.__data


class ProgressBar:
    def __init__(self,
                 max_iter: int = 1,
                 verbose: int = 1,
                 bar_nums: int = 20,
                 untrained_sign: str = '*',
                 trained_sign: str = '='):
        self.max_iter = max_iter
        self.verbose = verbose
        self._nums = bar_nums - 1
        self._untrained = untrained_sign
        self._trained = trained_sign
        self.iter = 0
        self.times = 0

    def update(self, n_iter: int = 1):
        self.iter += n_iter
        self.times += 1

    def get_bar(self) -> str:
        trained_ratio = self.iter / self.max_iter
        reached_bar_nums = round(trained_ratio * self._nums)
        unreached_bar_nums = self._nums - reached_bar_nums
        if self.verbose == 1:
            bar = reached_bar_nums * self._trained + '>' + unreached_bar_nums * self._untrained
        else:
            percent = str(round(trained_ratio * 100))
            bar = '{black} {percent:>{white}}%'.format(black="\033[40m%s\033[0m" % ' ' * reached_bar_nums,
                                                       percent=percent, white=unreached_bar_nums)
        return bar

    def console(self,
                verbose: int = 0,
                trained_time: float = 0.,
                batch_loss: float = 0.,
                batch_acc: float = 0.,
                validation_loss: float = None,
                validation_acc: float = None):

        if verbose == 0:
            return

        bar = self.get_bar()
        if verbose == 1:
            formated_trained_time = format_time(trained_time)
            formated_per_batch_time = format_time(trained_time / self.times)
            if validation_loss == 0. and validation_acc == 0.:
                print('\r {:d}/{:d} [{}] - {} - {}/batch -batch_loss: {:.4f} -batch_acc: {:.4f}'.format(self.iter,
                                                                                                        self.max_iter,
                                                                                                        bar,
                                                                                                        formated_trained_time,
                                                                                                        formated_per_batch_time,
                                                                                                        batch_loss,
                                                                                                        batch_acc),
                      flush=True, end='')
            else:
                print('\r {:d}/{:d} [{}] - {} - {}/batch'
                      ' -batch_loss: {:.4f} -batch_acc: {:.4f} -validation_loss: {:.4f} -validation_acc: {:.4f}'.format(
                    self.iter, self.max_iter, bar, formated_trained_time, formated_per_batch_time, batch_loss,
                    batch_acc, validation_loss, validation_acc), flush=True, end='')
        elif verbose == 2:
            sample_time = trained_time / self.iter
            eta = (self.max_iter - self.iter) * sample_time
            formated_eta = format_time(eta)
            if validation_loss == 0. and validation_acc == 0.:
                print('{} -ETA {} -batch_loss: {:.4f} -batch_acc: {:.4f}'.format(bar, formated_eta, batch_loss,
                                                                                 batch_acc))
            else:
                print(
                    '{} -ETA {} -batch_loss: {:.4f} -batch_acc: {:.4f} -validation_loss: {:.4f} -validation_acc: {:.4f}'.format(
                        bar, formated_eta, batch_loss, batch_acc, validation_loss, validation_acc))
        else:
            raise ValueError('Verbose only supports for 0, 1 and 2 ~')


def format_time(second_time: float) -> str:
    if second_time < 1:
        ms = second_time * 1000
        if ms < 1:
            us = second_time * 1000
            return '%dus' % us
        else:
            return '%dms' % ms
    second_time = round(second_time)
    if second_time > 3600:
        # hours
        h = second_time // 3600
        second_time = second_time % 3600
        # minutes
        m = second_time // 60
        second_time = second_time % 60
        return '%dh%dm%ds' % (h, m, second_time)
    elif second_time > 60:
        m = second_time // 60
        second_time = second_time % 60
        return '%dm%ds' % (m, second_time)
    else:
        return '%ds' % second_time


def gradient_check(inputs, target, layer, criterion, epsilon=1e-4):
    import numpy as np
    variables = list(layer.parameters())
    variables_mathematical_gradient_list = []
    with no_grad():
        for i in range(len(variables)):
            variable_shape = variables[i].shape
            variable_size = variables[i].numel()
            flat_variables = variables[i].eval.reshape(-1, 1)
            mathematical_gradient = np.zeros((variable_size, ))
            new_flat_add_variable = flat_variables.copy()
            new_flat_minus_variable = flat_variables.copy()
            for j in range(variable_size):
                new_flat_add_variable[j] += epsilon
                new_data = new_flat_add_variable.reshape(variable_shape)
                variables[i].eval = new_data
                layer.parameters(variables)
                out = layer(inputs)
                loss1 = criterion(out, target)
                # recover
                new_flat_add_variable[j] -= epsilon

                new_flat_minus_variable[j] -= epsilon
                new_data = new_flat_minus_variable.reshape(variable_shape)
                variables[i].eval = new_data
                layer.parameters(variables)
                out = layer(inputs)
                loss2 = criterion(out, target)
                # recover
                new_flat_minus_variable[j] += epsilon

                mathematical_gradient[j] = (loss1.eval - loss2.eval) / (2 * epsilon)
            mathematical_gradient = mathematical_gradient.reshape(variable_shape)
            variables_mathematical_gradient_list.append(mathematical_gradient)
    return variables_mathematical_gradient_list



from xs.optim.optimizer import *


class SGD(Optimizer):
    def step(self):
        for v in self._parameters:
            if v.requires_grad:
                v.eval -= self.lr * v.grad.eval
        super(SGD, self).step()


class Momentum(Optimizer):
    def __init__(self, parameters=None, lr=0.01, weight_decay=0., rho=0.9):
        self.rho = rho
        self.velocity = None
        super(Momentum, self).__init__(parameters=parameters, lr=lr, weight_decay=weight_decay)

    def step(self):
        if self.velocity is None:
            # initialize
            self.velocity = [GLOBAL.np.zeros_like(p.eval) for p in self._parameters]

        for i, (v, var) in enumerate(zip(self.velocity, self._parameters)):
            if var.requires_grad:
                v = self.rho * v + (1 - self.rho) * var.grad.eval
                var.eval -= self.lr * v
                self.velocity[i] = v

        super(Momentum, self).step()



from .base import *


class Input(Layer):
    def __init__(self, input_shape: Union[List, Tuple], **kwargs):
        super(Input, self).__init__(input_shape=input_shape, **kwargs)
        self._shape = input_shape

    def call(self, x: F.Tensor, *args, **kwargs) -> F.Tensor:
        self._data = x
        return self._data


class Reshape(Layer):
    def __init__(self, shape: Tuple, **kwargs):
        super().__init__(shape=shape, **kwargs)

    def call(self, x: F.Tensor, *args, **kwargs) -> F.Tensor:
        self._data = F.view(x, (-1, ) + self._shape, self._data)
        return self._data

    def compute_output_shape(self, input_shape: Union[List, Tuple] = None) -> Union[List, Tuple]:
        return self._shape


class ZeroPadding2D(Layer):
    def __init__(self, padding, **kwargs):
        self.padding = padding
        super(ZeroPadding2D, self).__init__(**kwargs)

    def call(self, x: F.Tensor, *args, **kwargs) -> F.Tensor:
        self._data = F.pad2d(x, self.padding, self._data)
        return self._data

    def compute_output_shape(self, input_shape: Union[List, Tuple] = None) -> Union[List, Tuple]:
        self._shape = (input_shape[0], input_shape[1] + 2 * self.padding[0], input_shape[2] + 2 * self.padding[1])
        return self._shape


class Add(Layer):
    def __call__(self, inbounds: List[Layer], *args, **kwargs):
        for inbound in inbounds:
            self._in_bounds.append(inbound)
            inbound.add_out_bounds(self)
            self._shape = inbound.shape
        return self

    def init_layer_out_tensor(self, x : F.Tensor = None):
        x = self._in_bounds[0].data if x is None else x
        if self._data is None or x.shape[0] > self._data.shape_capacity[0]:
            self._data = Zeros()((x.shape[0],) + self.shape, requires_grad=self.trainable)
            self._data.to('static')
            for in_bound in self._in_bounds:
                self._data.add_in_bounds(in_bound.data)
        elif x.shape[0] < self._data.shape_capacity[0]:
            if GLOBAL.TRAINING:
                self._data.slices(slice(None, x.shape[0], None))
            else:
                self._data = Zeros()((x.shape[0],) + self.shape, requires_grad=self.trainable)
                self._data.to('static')
                for in_bound in self._in_bounds:
                    self._data.add_in_bounds(in_bound.data)
        else:
            self._data.slices(slice(None, None, None))

    def forward(self, x: F.Tensor = None, *args, **kwargs) -> F.Tensor:
        self._data.zero_()
        for in_bound in self._in_bounds:
            GLOBAL.np.add(self._data.eval, in_bound.data.eval, out=self._data.eval)
            if GLOBAL.TRAINING and in_bound.data.requires_grad:
                initialize_ops_grad(in_bound.data)
            self._data.requires_grad = self._data.requires_grad or in_bound.data.requires_grad
        return self._data

    def compute_output_shape(self, input_shape: Union[List, Tuple] = None) -> Union[List, Tuple]:
        return self._shape

    def backward(self, gradients: F.Tensor = None):
        for in_bound in self._in_bounds:
            if in_bound.data.requires_grad:
                GLOBAL.np.add(in_bound.data.grad.eval, self._data.grad.eval, out=in_bound.data.grad.eval)
        self._data.zero_grad()



from xs.core import __global as GLOBAL
from xs.utils.common import ndarray


# base math operation
def add(a: ndarray, b: ndarray, out: ndarray = None):
    return GLOBAL.np.add(a, b, out=out)


def sub(a: ndarray, b: ndarray, out: ndarray = None):
    return GLOBAL.np.subtract(a, b, out=out)


def mul(a: ndarray, b: ndarray, out: ndarray = None):
    return GLOBAL.np.multiply(a, b, out=out)


def div(a: ndarray, b: ndarray, out: ndarray = None):
    return GLOBAL.np.divide(a, b, out=out)


def floor_div(a: ndarray, b: ndarray, out: ndarray = None):
    return GLOBAL.np.true_divide(a, b, out=out)


def mm(a: ndarray, b: ndarray, out: ndarray = None):
    return GLOBAL.np.dot(a, b, out=out)


def exp(x: ndarray, out: ndarray = None):
    return GLOBAL.np.exp(x, out=out)


def max(x: ndarray, axis: int = None, keepdims: bool = False, out: ndarray = None):
    return GLOBAL.np.max(x, axis=axis, keepdims=keepdims, out=out)


def maximum(x1: ndarray, x2: ndarray, out: ndarray = None):
    return GLOBAL.np.maximum(x1, x2, out=out)


def sum(x: ndarray, axis: int = None, keepdims: bool = False, out: ndarray = None):
    return GLOBAL.np.sum(x, axis, keepdims=keepdims, out=out)


def mean(x: ndarray, axis: int = None, keepdims: bool = False, out: ndarray = None):
    return GLOBAL.np.mean(x, axis, keepdims=keepdims, out=out)


def norm(x: ndarray, p: int = 2, axis: int = None, keepdims: bool = False, out: ndarray = None):
    x = GLOBAL.np.abs(x)
    x = GLOBAL.np.power(x, p)
    out = GLOBAL.np.sum(x, axis=axis, keepdims=keepdims, out=out)
    out = GLOBAL.np.power(out, 1 / p)
    return out


# base activation
def relu(x: ndarray, out: ndarray = None):
    return GLOBAL.np.maximum(0., x, out=out)


def sigmoid(x: ndarray, out: ndarray = None):
    out = GLOBAL.np.exp(-x, out=out)
    out = GLOBAL.np.add(1., out, out=out)
    return GLOBAL.np.divide(1., out, out=out)


def tanh(x: ndarray, out: ndarray = None):
    return GLOBAL.np.tanh(x, out=out)


def softmax(x: ndarray, out: ndarray = None):
    # more stable softmax
    out = GLOBAL.np.subtract(x, max(x, axis=-1, keepdims=True), out=out)
    GLOBAL.np.exp(out, out=out)
    GLOBAL.np.divide(out, sum(out, axis=-1, keepdims=True), out=out)
    return out


def log_softmax(x: ndarray, out: ndarray = None):
    out = softmax(x, out)
    out = GLOBAL.np.log(out, out=out)
    return out


# base nn function
def flatten(x: ndarray, start: int = 1):
    output_shape = tuple(x.shape[:start]) + (-1,)
    return GLOBAL.np.reshape(x, output_shape)


def long(data: ndarray):
    return data.astype(GLOBAL.np.int64)


def expand_as(inputs: ndarray, target: ndarray):
    new_axis_list = []
    inputs_idx = 0
    for i in range(target.ndim):
        if inputs_idx >= inputs.ndim:
            new_axis_list.append(i)
        elif inputs.shape[inputs_idx] == target.shape[i]:
            inputs_idx += 1
        else:
            new_axis_list.append(i)
    try:
        return GLOBAL.np.expand_dims(inputs, axis=new_axis_list)
    except TypeError:
        for a in new_axis_list:
            inputs = GLOBAL.np.expand_dims(inputs, axis=a)
        return inputs


def nll_loss(pred: ndarray, target: ndarray, reduction: str = 'mean', out: ndarray = None):
    to_sum_dim = GLOBAL.np.prod(GLOBAL.np.asarray(pred.shape[:-1])).item()
    log_probs = pred.reshape(-1, pred.shape[-1])
    y_flat = target.reshape(to_sum_dim).astype(GLOBAL.np.int)
    if reduction == 'sum':
        sum_val = -GLOBAL.np.sum(log_probs[GLOBAL.np.arange(to_sum_dim), y_flat])
        out = GLOBAL.np.multiply(-1, sum_val, out=out)
    elif reduction == 'mean':
        sum_val = -GLOBAL.np.sum(log_probs[GLOBAL.np.arange(to_sum_dim), y_flat])
        out = GLOBAL.np.divide(sum_val, pred.shape[0], out=out)
    else:
        out = GLOBAL.np.multiply(-1, log_probs[GLOBAL.np.arange(to_sum_dim), y_flat], out=out)
    out = GLOBAL.np.abs(out)
    return out


def bce_loss(pred: ndarray, target: ndarray, reduction: str = 'mean', out: ndarray = None):
    if reduction == 'sum':
        loss_val = GLOBAL.np.sum(GLOBAL.np.add(GLOBAL.np.multiply(target, GLOBAL.np.log(pred)),
                           GLOBAL.np.multiply(GLOBAL.np.subtract(1, target), GLOBAL.np.log(GLOBAL.np.subtract(1, pred)))))
        out = GLOBAL.np.multiply(-1, loss_val, out=out)
    elif reduction == 'mean':
        loss_val = GLOBAL.np.mean(GLOBAL.np.add(GLOBAL.np.multiply(target, GLOBAL.np.log(pred)),
                                 GLOBAL.np.multiply(GLOBAL.np.subtract(1, target), GLOBAL.np.log(GLOBAL.np.subtract(1, pred)))))
        out = GLOBAL.np.multiply(-1, loss_val, out=out)
    else:
        out = GLOBAL.np.add(GLOBAL.np.multiply(target, GLOBAL.np.log(pred)),
                           GLOBAL.np.multiply(GLOBAL.np.subtract(1, target), GLOBAL.np.log(GLOBAL.np.subtract(1, pred))), out=out)
        out = GLOBAL.np.multiply(-1, out, out=out)

    return out



from xs.utils.data import DataSet
from xs.utils.common import List
from xs.nn.initializers import Empty
from xs.core import __global as GLOBAL


class DataLoader:
    def __init__(self, dataset: DataSet, batch_size: int = None, shuffle: bool = False, seed: int = None):
        self.dataset = dataset
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.seed = 0 if seed is None else seed
        self.mini_batches = self.make_batches(self.dataset.datas, self.batch_size, self.seed, self.shuffle)
        self.sp = -1
        self.__tensor_pool = None
        self.__initialize_tensor_pool()

    def __iter__(self):
        return self

    def __initialize_tensor_pool(self):
        assert self.dataset.datas
        if self.batch_size > self.dataset.datas[0].shape[0]:
            self.__tensor_pool = [Empty()((self.dataset.datas[0].shape[0],) + self.dataset.datas[i].shape[1:],
                                          requires_grad=False) for i in range(len(self.dataset.datas))]
        else:
            self.__tensor_pool = [Empty()((self.batch_size,) + self.dataset.datas[i].shape[1:], requires_grad=False) for i in
                                  range(len(self.dataset.datas))]

    def __next__(self):
        self.sp += 1
        if self.sp >= len(self.mini_batches):
            self.seed += 1
            self.mini_batches = self.make_batches(self.dataset.datas, self.batch_size, self.seed, self.shuffle)
            self.sp = -1
            raise StopIteration
        for i, data in enumerate(self.mini_batches[self.sp]):
            if data.shape[0] < self.__tensor_pool[i].shape_capacity[0]:
                self.__tensor_pool[i].slices(slice(None, data.shape[0], None))
            elif self.__tensor_pool[i].shape_capacity[0] != self.__tensor_pool[i].shape[0] and data.shape[0] == self.__tensor_pool[i].shape_capacity[0]:
                self.__tensor_pool[i].slices(slice(None, None, None))
            self.__tensor_pool[i].data[:] = data
        return self.__tensor_pool[0] if len(self.__tensor_pool) == 1 else self.__tensor_pool

    def __len__(self):
        return len(self.mini_batches)

    def make_batches(self, datas: List[GLOBAL.np.ndarray], batch_size: int, seed: int, shuffle: bool = False):
        if batch_size is None:
            return [datas, ]
        GLOBAL.np.random.seed(seed)
        m = datas[0].shape[0]
        if shuffle:
            permutation = GLOBAL.np.random.permutation(m)
            datas = list(map(lambda x: GLOBAL.np.asarray(x)[permutation], datas))

        mini_batches = []
        complete_batch_nums = m // batch_size  # 完整的mini_batch个数
        for i in range(complete_batch_nums):
            mini_batch = list(map(lambda x: x[batch_size * i:batch_size * (i + 1)], datas))
            mini_batches.append(mini_batch)

        if m % batch_size != 0:
            mini_batches.append(list(map(lambda x: x[batch_size * complete_batch_nums:], datas)))
        return mini_batches



from xs.utils.common import GLOBAL, Union, List, ndarray
import xs.core.autograd
import xs.nn.functional as F


class Tensor(object):
    def __new__(cls, *args, **kwargs):
        if len(args) and isinstance(args[0], Tensor):
            return args[0]
        return super().__new__(cls)

    def __init__(self, data: Union[ndarray, int, float, List], requires_grad: bool = False, dtype: str = None,
                 name: str = None, **kwargs):

        if isinstance(data, Tensor):
            return
        self.__data = None
        self.__grad = None
        self.__in_bounds = None
        self.__out_bounds = None
        self.name = None
        self.grad_fn = None
        self.requires_grad = None
        self.__slice_items = None
        self.__static_graph_tensor = None
        # max storage tensor's shape
        self.__cache = None
        self.__retain_grad = None
        self.__is_leaf = None
        self.next_layers = None
        self.reset_(data, requires_grad, name, dtype, **kwargs)

    def reset_(self, data: Union[ndarray, int, float, List] = None, requires_grad: bool = False, name: str = None, dtype: str = None, **kwargs):
        if data is not None:
            self.__data = GLOBAL.np.asarray(data, dtype=dtype)
        self.__grad = kwargs.pop('grad', None)
        self.__in_bounds = []
        self.__out_bounds = []
        self.name = name
        self.grad_fn = None
        self.requires_grad = requires_grad
        self.__slice_items = kwargs.pop('slices', slice(None, None, None))
        self.__static_graph_tensor = kwargs.pop('static_graph_tensor', False)
        # max storage tensor's shape
        self.__cache = {}
        self.__retain_grad = kwargs.pop('retain_grad', False)
        self.__is_leaf = None
        self.next_layers = []

    def free_memory(self):
        del self.__data
        del self.__grad
        self.__data = None
        self.__grad = None

    @property
    def data(self):
        if self.__data is None:
            return None
        return Tensor(self.__data, slices=self.__slice_items, static_graph_tensor=self.__static_graph_tensor, grad=self.__grad)

    @data.setter
    def data(self, v):
        if isinstance(v, Tensor):
            self.__data[self.__slice_items] = v.eval
        else:
            raise TypeError('Unknown data type {}'.format(type(v)))

    @property
    def eval(self):
        return self.__data if self.__data.size == 1 else self.__data[self.__slice_items]

    @eval.setter
    def eval(self, v):
        self.__data[self.__slice_items] = GLOBAL.np.asarray(v)

    @property
    def shape_capacity(self):
        return self.__data.shape

    # truly shape
    @property
    def shape(self):
        return self.eval.shape if len(self.eval.shape) != 0 else (1, )

    @property
    def dtype(self):
        return str(self.__data.dtype)

    @property
    def device(self):
        import numpy
        return "cpu" if isinstance(self.__data, numpy.ndarray) else "cuda"

    def slices(self, slices: slice = None):
        if slices is None:
            return self.__slice_items
        assert isinstance(slices, slice)
        self.__slice_items = slices

    @property
    def is_leaf(self):
        return self.__is_leaf if self.__is_leaf is not None else len(self.__in_bounds) == 0

    @is_leaf.setter
    def is_leaf(self, flag: bool):
        self.__is_leaf = flag

    @property
    def grad(self):
        if self.__grad is not None:
            self.__grad.slices(self.__slice_items)
        return self.__grad

    @grad.setter
    def grad(self, g):
        self.__grad = Tensor(g)

    @property
    def cache(self):
        return self.__cache

    @property
    def is_static(self):
        return self.__static_graph_tensor

    @property
    def is_dynamic(self):
        return not self.__static_graph_tensor

    def retain_grad(self, retain: bool = None):
        if retain is None:
            return self.__retain_grad
        self.__retain_grad = retain

    def to(self, dst: str = 'static'):
        if dst == 'static':
            self.__static_graph_tensor = True
            return self
        elif dst == 'dynamic':
            self.__static_graph_tensor = False
            return self
        elif dst == 'cuda':
            if not GLOBAL.USE_CUDA:
                GLOBAL.USE_CUDA = True
                GLOBAL.np = __import__('cupy')
            ret = Tensor(GLOBAL.np.asarray(self.__data), requires_grad=self.requires_grad, dtype=self.dtype, name=self.name)
            if self.__grad is not None:
                ret.grad = Tensor(GLOBAL.np.asarray(self.__grad))
            return ret
        elif dst == "cpu":
            if GLOBAL.USE_CUDA:
                GLOBAL.USE_CUDA = False
                GLOBAL.np = __import__('numpy')
            ret = Tensor(self.__reform_ndarray(self.__data), requires_grad=self.requires_grad, dtype=self.dtype,
                         name=self.name)
            if self.__grad is not None:
                ret.grad = Tensor(self.__reform_ndarray(self.__grad))
            return ret

    def __reform_ndarray(self, target):
        return GLOBAL.np.asarray(target)

    def cuda(self):
        return self.to("cuda")

    def cuda_(self):
        if not GLOBAL.USE_CUDA:
            GLOBAL.USE_CUDA = True
            GLOBAL.np = __import__('cupy')
        self.__data = self.__reform_ndarray(self.__data)
        if self.__grad is not None:
            self.__grad = self.__reform_ndarray(self.__grad)

    def cpu(self):
        return self.to("cpu")

    def cpu_(self):
        if GLOBAL.USE_CUDA:
            GLOBAL.USE_CUDA = False
            GLOBAL.np = __import__('numpy')
        self.__data = self.__reform_ndarray(self.__data)
        if self.__grad is not None:
            self.__grad = self.__reform_ndarray(self.__grad)

    def add_in_bounds(self, *ins):
        for i in ins:
            if isinstance(i, Tensor) or i is None:
                self.__in_bounds.append(i)
            else:
                raise TypeError('Not a Tensor')

    def add_out_bounds(self, *outs):
        for o in outs:
            if isinstance(o, Tensor):
                self.__out_bounds.append(o)
            else:
                raise TypeError('Not a Tensor')

    def requires_grad_(self, flag: bool):
        self.requires_grad = flag

    @property
    def out_bounds(self):
        return self.__out_bounds

    @property
    def in_bounds(self):
        return self.__in_bounds

    def numpy(self):
        import numpy
        return numpy.asarray(self.eval)

    def cupy(self):
        import cupy
        return cupy.asarray(self.eval)

    def __getitem__(self, item):
        ret = Tensor(self.data[item], requires_grad=self.requires_grad, slices=item)
        ret.add_out_bounds(self)
        if self.__grad is not None:
            ret.grad = self.__grad[item]
        return ret

    def __setitem__(self, key, value):
        if self.__slice_items is None:
            self.__data[key] = GLOBAL.np.asarray(value, dtype=self.__data.dtype)
        else:
            self.__data[self.__slice_items][key] = GLOBAL.np.asarray(value, dtype=self.__data.dtype)

    def __repr__(self) -> str:
        requires_grad = f", requires_grad={self.requires_grad}" if self.requires_grad else ""
        grad_fn = f', grad_fn=<{self.grad_fn.__name__}>' if self.grad_fn else ""
        info = f'{self.__class__.__name__}({self.__data}{requires_grad}{grad_fn})'
        return info

    def __str__(self) -> str:
        requires_grad = f", requires_grad={self.requires_grad}" if self.requires_grad else ""
        grad_fn = f', grad_fn=<{self.grad_fn.__name__}>' if self.grad_fn else ""
        info = f'{self.__class__.__name__}({self.__data}{requires_grad}{grad_fn})'
        return info

    # overload operators
    def __add__(self, other):
        other = Tensor(other)
        return F.add(self, other)

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        other = Tensor(other)
        return F.sub(self, other)

    def __rsub__(self, other):
        return self.__sub__(other)

    def __mul__(self, other):
        other = Tensor(other)
        return F.mul(self, other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __matmul__(self, other):
        other = Tensor(other)
        return F.mm(self, other)

    def __rmatmul__(self, other):
        return self.__matmul__(other)

    def __truediv__(self, other):
        other = Tensor(other)
        return F.truediv(self, other)

    def __floordiv__(self, other):
        other = Tensor(other)
        return F.floordiv(self, other)

    def __pow__(self, power, modulo=None):
        return F.pow(self, exp=power)

    def dot(self, other):
        return self.__matmul__(other)

    def mm(self, other):
        return self.__matmul__(other)

    def numel(self):
        return self.__data.size

    def item(self):
        assert self.__data.size == 1
        return self.__data.item()

    def long_(self):
        self.__data = self.__data.astype(GLOBAL.np.int64)

    def zero_(self):
        self.__data = GLOBAL.np.zeros_like(self.__data)

    def ones_(self):
        self.__data = GLOBAL.np.ones_like(self.__data)

    def fill_(self, value):
        self.__data.fill(value)

    def full(self, value):
        return Tensor(GLOBAL.np.full(self.shape), value)

    def zero_grad(self):
        if self.__grad is None:
            self.__grad = Tensor(GLOBAL.np.zeros(self.shape, dtype=self.dtype))
        else:
            self.__grad.zero_()

    def sum(self, axis: int = None, keepdims: bool = False):
        return F.sum(self, axis, keepdims)

    def mean(self, axis: int = None, keepdims: bool = False):
        return F.mean(self, axis, keepdims)

    def l2norm(self):
        return F.norm(self)

    def t(self):
        return F.transpose(self)

    def sigmoid(self):
        return F.sigmoid(self)

    def view(self, *shapes):
        return F.view(self, shapes)

    def size(self, *dims):
        if len(dims) == 0:
            return self.shape
        elif len(dims) == 1:
            return self.shape[dims[0]]
        return (self.shape[d] for d in dims)

    def backward(self, gradients=None, retain_graph=False):
        if self.grad_fn is None:
            raise ValueError('can not solve grad on %s' % self)
        if gradients is not None:
            assert gradients.size == self.__data.size and gradients.shape == self.shape
            self.__grad = Tensor(GLOBAL.np.asarray(gradients, dtype=self.dtype))
        else:
            if self.__data.size == 1:
                if self.__grad is None:
                    self.__grad = Tensor(GLOBAL.np.array(1., dtype=self.dtype))
            else:
                raise ValueError('grad can be implicitly created only for scalar outputs')
        GLOBAL.OUTPUTS = self
        # if GLOBAL.INPUTS is not None:
        xs.core.autograd.backward(self, GLOBAL.INPUTS, retain_graph)


class Parameter(Tensor):
    def __new__(cls, *args, **kwargs):
        if len(args) and isinstance(args[0], Parameter):
            return args[0]
        return super().__new__(cls)

    def __init__(self, t: Union[Tensor, ndarray, int, float], **kwargs):
        if isinstance(t, Tensor):
            super(Parameter, self).__init__(t.eval, t.requires_grad, t.dtype, t.name,
                                            **kwargs)
        else:
            super(Parameter, self).__init__(data=t, requires_grad=True)



from xs.utils.common import copy, GLOBAL
from xs.nn.initializers import Zeros


class Optimizer:
    def __init__(self, parameters=None, lr=0.01, weight_decay=0.):
        self._parameters = parameters
        self.lr = lr
        self.weight_decay = weight_decay
        self.iterations = 0

    def zero_grad(self):
        for v in self._parameters:
            if v.grad is None:
                v.zero_grad()
            else:
                v.grad.zero_()

    def step(self):
        self.iterations += 1



import xs.core.base
from xs.utils.common import *


class Initializer:
    def __init__(self, seed=None):
        if seed is not None:
            GLOBAL.np.random.seed(seed)

    def decompose_size(self, shape):
        shape = GLOBAL.np.array(shape)
        if shape.ndim == 2:
            fan_in, fan_out = shape
        elif shape.ndim == 4 or shape.ndim == 5:
            respective_field_size = GLOBAL.np.prod(shape[2:])
            fan_in = shape[1] * respective_field_size
            fan_out = shape[0] * respective_field_size
        else:
            fan_in = fan_out = int(GLOBAL.np.sqrt(GLOBAL.np.prod(shape)))
        return fan_in, fan_out


class Uniform(Initializer):
    def __init__(self, scale=0.05, seed=None):
        self.scale = scale
        super(Uniform, self).__init__(seed)

    def __call__(self, shape, **kwargs):
        return xs.core.base.Tensor(GLOBAL.np.random.uniform(-self.scale, self.scale, size=shape), **kwargs)


class LecunUniform(Initializer):
    def __init__(self, seed=None):
        super(LecunUniform, self).__init__(seed)

    def __call__(self, shape, **kwargs):
        fan_in, fan_out = self.decompose_size(shape)
        return Uniform(GLOBAL.np.sqrt(3. / fan_in))(shape, **kwargs)


class GlorotUniform(Initializer):
    def __init__(self, seed=None):
        super(GlorotUniform, self).__init__(seed)

    def __call__(self, shape, **kwargs):
        fan_in, fan_out = self.decompose_size(shape)
        return Uniform(GLOBAL.np.sqrt(6. / (fan_in + fan_out)))(shape, **kwargs)


class HeUniform(Initializer):
    def __init__(self, seed=None):
        super(HeUniform, self).__init__(seed)

    def __call__(self, shape, **kwargs):
        fan_in, fan_out = self.decompose_size(shape)
        return Uniform(GLOBAL.np.sqrt(6. / fan_in))(shape, **kwargs)


class Normal(Initializer):
    def __init__(self, std=0.1, mean=0.0, seed=None):
        self.std = std
        self.mean = mean
        super(Normal, self).__init__(seed)

    def __call__(self, shape, **kwargs):
        return xs.core.base.Tensor(GLOBAL.np.random.normal(loc=self.mean, scale=self.std, size=shape), **kwargs)


class LecunNormal(Initializer):
    def __init__(self, seed=None):
        super(LecunNormal, self).__init__(seed)

    def __call__(self, shape, **kwargs):
        fan_in, fan_out = self.decompose_size(shape)
        return Normal(GLOBAL.np.sqrt(1. / fan_in))(shape, **kwargs)


class GlorotNormal(Initializer):
    def __init__(self, seed=None):
        super(GlorotNormal, self).__init__(seed)

    def __call__(self, shape, **kwargs):
        fan_in, fan_out = self.decompose_size(shape)
        return Normal(GLOBAL.np.sqrt(2. / (fan_in + fan_out)))(shape, **kwargs)


class HeNormal(Initializer):
    def __init__(self, seed=None):
        super(HeNormal, self).__init__(seed)

    def __call__(self, shape, **kwargs):
        fan_in, fan_out = self.decompose_size(shape)
        return Normal(GLOBAL.np.sqrt(2. / fan_in))(shape, **kwargs)


class Orthogonal(Initializer):
    def __init__(self, gain=1.0, seed=None):
        if gain == 'relu':
            gain = GLOBAL.np.sqrt(2)
        self.gain = gain
        super(Orthogonal, self).__init__(seed)

    def __call__(self, shape, **kwargs):
        shape = GLOBAL.np.array(shape)
        flat_shape = (shape[0].tolist(), GLOBAL.np.prod(shape[1:]).tolist())
        a = Normal(1.)(flat_shape, **kwargs)
        u, _, v = GLOBAL.np.linalg.svd(a.data, full_matrices=False)
        q = u if u.shape == flat_shape else v
        q = q.reshape(shape.tolist())
        a.data.data = self.gain * q
        return a


class Zeros(Initializer):
    def __call__(self, shape, **kwargs):
        return xs.core.base.Tensor(GLOBAL.np.zeros(shape), **kwargs)


class Ones(Initializer):
    def __call__(self, shape, **kwargs):
        return xs.core.base.Tensor(GLOBAL.np.ones(shape), **kwargs)


class Matrix(Initializer):
    def __init__(self, value: float, seed=None):
        super().__init__(seed=seed)
        self.value = value

    def __call__(self, shape: Tuple, **kwargs):
        return xs.core.base.Tensor(GLOBAL.np.ones(shape) * self.value, **kwargs)


class RandN(Initializer):
    def __call__(self, shape: Tuple, **kwargs):
        return xs.core.base.Tensor(GLOBAL.np.random.randn(*shape), **kwargs)


class Rand(Initializer):
    def __call__(self, shape: Tuple, **kwargs):
        return xs.core.base.Tensor(GLOBAL.np.random.rand(*shape), **kwargs)


class RandInt(Initializer):
    def __call__(self, low, high=None, shape=None, **kwargs):
        return xs.core.base.Tensor(GLOBAL.np.random.randint(low=low, high=high, size=shape), **kwargs)


def matrix(*shape, value=0., **kwargs):
    requires_grad = kwargs.pop('requires_grad', False)
    return Matrix(value=value)(shape, requires_grad=requires_grad, **kwargs)


def ones(*shape, **kwargs):
    requires_grad = kwargs.pop('requires_grad', False)
    return Ones()(shape, requires_grad=requires_grad, **kwargs)


def ones_like(a, **kwargs):
    requires_grad = kwargs.pop('requires_grad', a.requires_grad)
    return xs.core.base.Tensor(GLOBAL.np.ones_like(a.data), requires_grad=requires_grad, **kwargs)


def zeros(*shape, **kwargs):
    requires_grad = kwargs.pop('requires_grad', False)
    return Zeros()(shape, requires_grad=requires_grad, **kwargs)


def zeros_like(a, **kwargs):
    requires_grad = kwargs.pop('requires_grad', a.requires_grad)
    return xs.core.base.Tensor(GLOBAL.np.zeros_like(a.data), requires_grad=requires_grad, **kwargs)


def rand(*shape, **kwargs):
    requires_grad = kwargs.pop('requires_grad', False)
    return Rand()(shape, requires_grad=requires_grad, **kwargs)


def randn(*shape, **kwargs):
    requires_grad = kwargs.pop('requires_grad', False)
    return RandN()(shape, requires_grad=requires_grad, **kwargs)


def randint(low, high=None, shape=None, **kwargs):
    requires_grad = kwargs.pop('requires_grad', False)
    return RandInt()(low, high, shape, requires_grad=requires_grad, **kwargs)


def tensor(data, **kwargs):
    requires_grad = kwargs.pop('requires_grad', False)
    return xs.core.base.Tensor(data=data, requires_grad=requires_grad, **kwargs)


def manual_seed_all(seeds=None):
    GLOBAL.np.random.seed(seeds)


def get_initializer(initializer):
    if initializer.__class__.__name__ == 'str':
        initializer = initializer.lower()
        if initializer == 'zeros':
            return Zeros()
        elif initializer == 'ones':
            return Ones()
        elif initializer in ['heuniform', 'he_uniform']:
            return HeUniform()
        elif initializer == 'uniform':
            return Uniform()
        elif initializer == 'normal':
            return Normal()
        elif initializer in ['henormal', 'he_normal']:
            return HeNormal()
        elif initializer in ['lecunnormal', 'lecun_normal']:
            return LecunNormal()
        elif initializer in ['lecununiform', 'lecun_uniform']:
            return LecunUniform()
        elif initializer == 'orthogonal':
            return Orthogonal()
        elif initializer in ['glorotnoraml', 'glorot_normal']:
            return GlorotNormal()
        elif initializer in ['glorotuniform', 'glorot_uniform']:
            return GlorotUniform()

    elif isinstance(initializer, Initializer):
        return copy.deepcopy(initializer)
    else:
        raise TypeError('unknown initialization type!')




from xs.layers.activators import get_activator
from .base import *


class Dense(Layer):
    def __init__(self, out_features: int, activation: str = None, use_bias: bool = True, kernel_initializer: str = 'normal',
                 bias_initializer: str = 'zeros', **kwargs):
        self.out_features = out_features
        self.use_bias = use_bias
        self.kernel_initializer = get_initializer(kernel_initializer)
        self.bias_initializer = get_initializer(bias_initializer)
        self.activation = get_activator(activation) if activation is not None else None
        super().__init__(**kwargs)

    @property
    def data(self):
        return self._data if self.activation is None else self.activation.data

    @data.setter
    def data(self, v):
        if self.activation is None:
            self._data = v
        else:
            self.activation.data = v

    def compute_output_shape(self, input_shape: Union[List, Tuple] = None) -> Union[List, Tuple]:
        return (self.out_features, )

    def init_params(self, input_shape=None, *args):
        if input_shape is not None:
            self._input_shape = input_shape
        weight = F.Parameter(self.kernel_initializer(self._input_shape + (self.out_features, ), requires_grad=True))
        self._parameters.append(weight)
        if self.use_bias:
            bias = F.Parameter(self.bias_initializer((1, self.out_features), requires_grad=True))
            self._parameters.append(bias)

    def init_layer_out_tensor(self, x: F.Tensor = None):
        super(Dense, self).init_layer_out_tensor(x)
        if self.activation is not None:
            self.activation.compute_output_shape(self.out_features)
            self.activation.init_layer_out_tensor(self._data)

    def call(self, x: F.Tensor, *args, **kwargs) -> F.Tensor:
        if self.use_bias:
            weight, bias = self._parameters
            self._data = F.addmm(bias, x, weight, self._data)
        else:
            weight,  = self._parameters
            self._data = F.mm(x, weight, self._data)
        if self.activation is not None:
            return self.activation.forward(self._data)
        return self._data

    def backward(self, gradients: F.Tensor = None):
        if self.activation is not None:
            self.activation.backward()
        super().backward()


class Flatten(Layer):
    def __init__(self, start_dim=1, **kwargs):
        if start_dim < 1:
            raise ValueError('start_dim must be > 0')
        self.start_dim = start_dim
        super(Flatten, self).__init__(**kwargs)

    def compute_output_shape(self, input_shape=None):
        assert len(input_shape) >= 3
        flatten_shape = reduce(lambda x, y: x * y, input_shape[self.start_dim - 1:])
        output_shape = input_shape[: self.start_dim - 1] + (flatten_shape,)
        return output_shape

    def call(self, x: F.Tensor, *args, **kwargs) -> F.Tensor:
        self._data = F.flatten(x, self.start_dim,  self._data)
        return self._data



from xs.utils.toolkit import gradient_check, SummaryProfile
from xs.nn.rng import manual_seed_all, randn, tensor, zeros, ones, rand, randint
import xs.nn


# dtype
float16 = nn.GLOBAL.np.dtype(nn.GLOBAL.np.float16)
float32 = nn.GLOBAL.np.dtype(nn.GLOBAL.np.float32)
float64 = nn.GLOBAL.np.dtype(nn.GLOBAL.np.float64)
int8 = nn.GLOBAL.np.dtype(nn.GLOBAL.np.int8)
uint8 = nn.GLOBAL.np.dtype(nn.GLOBAL.np.uint8)
int16 = nn.GLOBAL.np.dtype(nn.GLOBAL.np.int16)
uint16 = nn.GLOBAL.np.dtype(nn.GLOBAL.np.uint16)
int32 = nn.GLOBAL.np.dtype(nn.GLOBAL.np.int32)
uint32 = nn.GLOBAL.np.dtype(nn.GLOBAL.np.uint32)
int64 = nn.GLOBAL.np.dtype(nn.GLOBAL.np.int64)
uint64 = nn.GLOBAL.np.dtype(nn.GLOBAL.np.uint64)
bool = nn.GLOBAL.np.dtype(nn.GLOBAL.np.bool)


def cuda_available():
    try:
        import cupy
    except ImportError:
        return False
    return True
